import json
import uuid
import time
import sys
import os
import shutil

from google.protobuf.empty_pb2 import Empty
from blessed import Terminal  # type: ignore
import click
import docker  # type: ignore
import grpc  # type: ignore

from cicadad.core import containers
from cicadad.protos import hub_pb2_grpc, hub_pb2
from cicadad.util import constants
from cicadad import templates as templates_module


@click.group()
@click.option("--debug", type=bool, default=False, is_flag=True)
@click.pass_context
def cli(ctx, debug):
    ctx.ensure_object(dict)

    ctx.obj["DEBUG"] = debug


@cli.command()
@click.argument("build-path", type=str, default=".")
@click.pass_context
def init(ctx, build_path):
    # Create dockerfile from template if not exists
    if not os.path.exists(os.path.join(build_path, "Dockerfile")):

        if os.getenv("ENV") == "local":
            dockerfile_path = os.path.join(
                os.path.dirname(templates_module.__file__), "local.dockerfile"
            )
        elif os.getenv("ENV") == "dev":
            dockerfile_path = os.path.join(
                os.path.dirname(templates_module.__file__), "dev.dockerfile"
            )
        else:
            dockerfile_path = os.path.join(
                os.path.dirname(templates_module.__file__), "Dockerfile"
            )

        # NOTE: Does not build subdirs to copy file
        shutil.copyfile(dockerfile_path, os.path.join(build_path, "Dockerfile"))

        if ctx.obj["DEBUG"]:
            click.echo("Added Dockerfile")

    # Create engine file from template if not exists
    if not os.path.exists(os.path.join(build_path, "test.py")):

        test_file = os.path.join(
            os.path.dirname(templates_module.__file__),
            "test.py",
        )

        shutil.copyfile(test_file, os.path.join(build_path, "test.py"))

        if ctx.obj["DEBUG"]:
            click.echo("Added test.py")


@cli.command()
@click.pass_context
@click.option("--network", type=str, default=constants.DEFAULT_DOCKER_NETWORK)
@click.option("--create-network/--no-create-network", default=True)
def start_cluster(ctx, network, create_network):
    # FEATURE: more options for docker client
    docker_client = docker.from_env()

    containers.configure_docker_network(docker_client, network, create_network)

    redis_container = containers.docker_redis_up(docker_client, network)

    if ctx.obj["DEBUG"]:
        click.echo(f"Created Redis: {redis_container.id}")

    datastore_client = containers.docker_datastore_client_up(docker_client, network)

    if ctx.obj["DEBUG"]:
        click.echo(f"Created Datastore Client: {datastore_client.id}")

    container_service = containers.docker_container_service_up(docker_client, network)

    if ctx.obj["DEBUG"]:
        click.echo(f"Created Container Service: {container_service.id}")


@cli.command()
@click.pass_context
def stop_cluster(ctx):
    docker_client = docker.from_env()

    containers.docker_container_service_down(docker_client)

    if ctx.obj["DEBUG"]:
        click.echo("Stopped Container Service")

    containers.docker_datastore_client_down(docker_client)

    if ctx.obj["DEBUG"]:
        click.echo("Stopped Datastore Client")

    containers.docker_redis_down(docker_client)

    if ctx.obj["DEBUG"]:
        click.echo("Stopped Redis")

    containers.clean_docker_containers(
        docker_client,
        [
            "cicada-distributed-test",
            "cicada-distributed-scenario",
            "cicada-distributed-user",
        ],
    )

    # NOTE: remove network?


@cli.command()
@click.pass_context
@click.option("--image", type=str, required=False)
@click.option("--build-path", type=str, default=".")
@click.option("--dockerfile", type=str, default="Dockerfile")
@click.option("--network", type=str, default=constants.DEFAULT_DOCKER_NETWORK)
@click.option("--tag", "-t", type=str, multiple=True, default=[])
@click.option("--no-exit-unsuccessful", is_flag=True)
def run(ctx, image, build_path, dockerfile, network, tag, no_exit_unsuccessful):
    # FIXME: move to function with dependencies removed
    docker_client = docker.from_env()
    term = Terminal()

    if image:
        image_id = image
    else:
        image_id = containers.build_docker_image(
            client=docker_client,
            path=build_path,
            dockerfile=dockerfile,
        )

    name = f"cicada-test-{str(uuid.uuid4())[:8]}"

    click.echo(
        term.bold
        + term.center(f" Starting Test: {name} ", fillchar="=")
        + "\n"
        + term.normal
    )

    # FEATURE: override datastore client/container service address from CLI
    args = containers.DockerServerArgs(
        image=image_id,
        name=name,
        command=[
            "run-test",
            "--image",
            image_id,
            "--network",
            network,
        ],
        labels=["cicada-distributed-test"],
        # env: Dict[str, str]={}
        # volumes: Optional[List[Volume]]
        host_port=8282,
        container_port=50051,
        network=network,
        # create_network: bool=True
    )

    test_container = containers.create_docker_container(docker_client, args)

    # FEATURE: Error if cluster is not up

    try:
        # FIXME: move to function
        with grpc.insecure_channel("[::]:8282") as channel:
            stub = hub_pb2_grpc.HubStub(channel)

            # FIXME: integrate with util.backoff
            ready = False
            tries = 0
            period = 2

            while not ready and tries < 3:
                try:
                    response = stub.Healthcheck(Empty())
                    ready = response.ready
                except grpc.RpcError as e:
                    ready = False

                    if ctx.obj["DEBUG"]:
                        click.echo(e)

                    time.sleep(period)

                    tries += 1
                    period *= 2

            if not ready:
                logs = containers.container_logs(test_container)

                raise RuntimeError(f"Could not start test:\n{logs}")

        # FIXME: move to function
        context = {}
        passed = []
        failed = []

        with grpc.insecure_channel("[::]:8282") as channel:
            stub = hub_pb2_grpc.HubStub(channel)
            msg = hub_pb2.RunTestRequest(tags=tag)

            for status in stub.Run(msg):
                # HACK: maybe change this to send back only the results for scenario instead of whole context

                if status.type == "SCENARIO_FINISHED":
                    context = json.loads(status.context)

                    result = context[status.scenario]

                    if result["exception"] is not None:
                        click.echo(
                            term.bold
                            + term.center(
                                f" {status.scenario}: "
                                + term.red
                                + "Failed "
                                + term.normal
                                + term.bold,
                                fillchar="-",
                            )
                            + "\n"
                            + term.normal
                        )

                        click.echo(f"Exception: {result['exception']}")
                        click.echo("\n")

                        failed.append(status.scenario)
                    else:
                        click.echo(
                            term.bold
                            + term.center(
                                f" {status.scenario}: "
                                + term.green
                                + "Passed "
                                + term.normal
                                + term.bold,
                                fillchar="-",
                            )
                            + "\n"
                            + term.normal
                        )

                        passed.append(status.scenario)

                    if result["output"] is not None:
                        click.echo(f"Result: {result['output']}")
                        click.echo("\n")

                    if result["logs"] and (
                        result["exception"] is not None or ctx.obj["DEBUG"]
                    ):
                        click.echo("Logs:")
                        click.echo(result["logs"])
                        click.echo("\n")
                else:
                    click.echo(
                        term.bold
                        + term.center(f" {status.message} ", fillchar="-")
                        + "\n"
                        + term.normal
                    )
    finally:
        if ctx.obj["DEBUG"]:
            click.echo("Cleaning Test Runners")

        containers.stop_docker_container_by_name(docker_client, test_container.id)
        containers.clean_docker_containers(docker_client, "cicada-distributed-test")

        if ctx.obj["DEBUG"]:
            click.echo("Cleaned Test Runners")
            click.echo("Cleaning Scenarios")

        containers.clean_docker_containers(docker_client, "cicada-distributed-scenario")

        if ctx.obj["DEBUG"]:
            click.echo("Cleaned Scenarios")
            click.echo("Cleaning Users")

        containers.clean_docker_containers(docker_client, "cicada-distributed-user")

        if ctx.obj["DEBUG"]:
            click.echo("Cleaned Users")

    # FEATURE: report results in static site
    click.echo(
        term.bold + term.center(" Test Complete ", fillchar="=") + "\n" + term.normal
    )

    if len(passed) > 0:
        click.echo(term.bold + "Passed:\n" + term.normal)

        for scenario in passed:
            click.echo("* " + term.green + scenario + term.normal)

    if len(failed) > 0:
        click.echo(term.bold + "Failed:\n" + term.normal)

        for scenario in failed:
            click.echo("* " + term.red + scenario + term.normal)

    click.echo(
        term.bold
        + "\n"
        + term.center(
            " "
            + term.green
            + f"{len(passed)} passed"
            + term.normal
            + term.bold
            + ", "
            + term.red
            + f"{len(failed)} failed"
            + " "
            + term.normal
            + term.bold,
            fillchar="=",
        )
        + "\n"
        + term.normal
    )

    if no_exit_unsuccessful:
        sys.exit(0)
    elif failed != []:
        sys.exit(1)


# init project (load base files)
# package runner for docker

if __name__ == "__main__":
    cli()



from typing import Any, Iterable, List, Optional
from datetime import datetime
import pickle  # nosec
import json

import grpc  # type: ignore

# from cassandra.cluster import Cluster, Session
# from cassandra.auth import PlainTextAuthProvider
from pydantic import BaseModel
from google.protobuf import wrappers_pb2

from cicadad.protos import datastore_pb2, datastore_pb2_grpc
from cicadad.util.constants import DEFAULT_DATASTORE_ADDRESS


class Result(BaseModel):
    id: Optional[str]
    output: Optional[Any]
    exception: Optional[Any]
    logs: Optional[str]
    timestamp: Optional[datetime]
    time_taken: Optional[int]

    class Config:
        json_encoders = {
            Exception: lambda e: str(e),
        }


def add_user_result(
    user_id: str, result: Result, address: str = DEFAULT_DATASTORE_ADDRESS
):
    with grpc.insecure_channel(address) as channel:
        stub = datastore_pb2_grpc.DatastoreStub(channel)
        request = datastore_pb2.AddUserResultRequest(
            userID=user_id, result=pickle.dumps(result)
        )

        stub.AddUserResult(request)


def set_scenario_result(
    scenario_id: str,
    output: Any,
    exception: Any,
    logs: str,
    time_taken: float,
    address: str = DEFAULT_DATASTORE_ADDRESS,
):
    with grpc.insecure_channel(address) as channel:
        stub = datastore_pb2_grpc.DatastoreStub(channel)
        request = datastore_pb2.SetScenarioResultRequest(
            scenarioID=scenario_id,
            output=wrappers_pb2.StringValue(value=json.dumps(output)),
            exception=wrappers_pb2.StringValue(
                value=json.dumps(str(exception) if exception is not None else None)
            ),
            logs=logs,
            timeTaken=time_taken,
        )

        stub.SetScenarioResult(request)


def move_user_results(
    user_ids: Iterable[str], address: str = DEFAULT_DATASTORE_ADDRESS
) -> List[Result]:
    with grpc.insecure_channel(address) as channel:
        stub = datastore_pb2_grpc.DatastoreStub(channel)
        request = datastore_pb2.MoveUserResultsRequest(userIDs=user_ids)

        response = stub.MoveUserResults(request)

        return [pickle.loads(result) for result in response.results]  # nosec


def move_scenario_result(
    scenario_id: str, address: str = DEFAULT_DATASTORE_ADDRESS
) -> Optional[dict]:
    with grpc.insecure_channel(address) as channel:
        try:
            stub = datastore_pb2_grpc.DatastoreStub(channel)
            request = datastore_pb2.MoveScenarioResultRequest(
                scenarioID=scenario_id,
            )

            response = stub.MoveScenarioResult(request)

            return {
                "id": response.id,
                "output": json.loads(response.output.value),
                "exception": json.loads(response.exception.value),
                "logs": response.logs,
                "timestamp": response.timestamp,
                "time_taken": response.timeTaken,
            }
        except grpc.RpcError as err:
            if err.code() == grpc.StatusCode.NOT_FOUND:
                return None
            else:
                raise err


def distribute_work(
    work: int, user_ids: List[str], address: str = DEFAULT_DATASTORE_ADDRESS
):
    with grpc.insecure_channel(address) as channel:
        stub = datastore_pb2_grpc.DatastoreStub(channel)
        request = datastore_pb2.DistributeWorkRequest(work=work, userIDs=user_ids)

        stub.DistributeWork(request)


def get_work(user_id: str, address: str = DEFAULT_DATASTORE_ADDRESS) -> int:
    with grpc.insecure_channel(address) as channel:
        stub = datastore_pb2_grpc.DatastoreStub(channel)
        request = datastore_pb2.GetUserWorkRequest(userID=user_id)

        response = stub.GetUserWork(request)

        return response.work


# def get_session(host: str = "0.0.0.0", password: str = "cassandra"):
#     cluster = create_datastore_cluster(host, password)

#     return cluster.connect()


# def setup_datastore(session: Session):
#     create_cicada_keyspace(session)

#     # NOTE: idk if this should be more configurable
#     session.set_keyspace("cicada")

#     create_user_results_table(session)
#     create_scenario_results_table(session)

#     return session


# def create_datastore_cluster(host: str, password: str):
#     # FEATURE: generate password for service creation
#     auth_provider = PlainTextAuthProvider(
#         username="cassandra",
#         password=password,
#     )

#     return Cluster([host], port=9042, auth_provider=auth_provider)


# def create_cicada_keyspace(session: Session):
#     return session.execute(
#         """
#         CREATE KEYSPACE IF NOT EXISTS cicada
#         WITH REPLICATION = { 'class' : 'SimpleStrategy', 'replication_factor' : 3 }
#         """,
#     )


# def create_user_results_table(session: Session):
#     return session.execute(
#         """
#         CREATE TABLE IF NOT EXISTS user_results
#         (
#             id UUID,
#             name TEXT,
#             output BLOB,
#             exception BLOB,
#             logs TEXT,
#             timestamp TIMESTAMP,
#             PRIMARY KEY (name, id)
#         )
#         """
#     )


# def create_scenario_results_table(session: Session):
#     return session.execute(
#         """
#         CREATE TABLE IF NOT EXISTS scenario_results
#         (
#             name TEXT PRIMARY KEY,
#             output BLOB,
#             exception BLOB,
#             logs TEXT,
#             timestamp TIMESTAMP
#         )
#         """
#     )


# def put_user_result(session: Session, name: str, result: Result):
#     return session.execute(
#         "INSERT INTO user_results (id, name, output, exception, logs, timestamp) VALUES (%s, %s, %s, %s, %s, %s)",
#         (
#             uuid.UUID(result.id),
#             name,
#             pickle.dumps(result.output),
#             pickle.dumps(result.exception),
#             result.logs,
#             result.timestamp,
#         ),
#     )


# def get_user_results(session: Session, name: str):
#     # NOTE: may be more performant to create a table for each set of user results
#     rows = session.execute(
#         """
#         SELECT * FROM user_results
#         WHERE name=%s
#         ALLOW FILTERING
#         """,
#         (name,),
#     ).all()

#     return [
#         Result(
#             id=str(row.id),
#             output=pickle.loads(row.output),
#             exception=pickle.loads(row.exception),
#             timestamp=row.timestamp,
#         )
#         for row in rows
#     ]


# def clear_user_results(session: Session, name: str):
#     return session.execute(
#         """
#         DELETE FROM user_results WHERE name=%s
#         """,
#         (name,),
#     )


# def put_scenario_result(session: Session, name: str, result: Result):
#     return session.execute(
#         "INSERT INTO scenario_results (name, output, exception, logs, timestamp) VALUES (%s, %s, %s, %s, %s)",
#         (
#             name,
#             pickle.dumps(result.output),
#             pickle.dumps(result.exception),
#             result.logs,
#             result.timestamp,
#         ),
#     )


# def get_scenario_result(session: Session, name: str):
#     row = session.execute(
#         "SELECT * FROM scenario_results WHERE name=%s ALLOW FILTERING", (name,)
#     ).one()

#     # HACK: will need to seperate pickling/unpickling from DB call if using
#     # API to consolidate shared logic
#     return Result(
#         output=pickle.loads(row.output),
#         exception=pickle.loads(row.exception),
#         timestamp=row.timestamp,
#     )


# def clear_scenario_results(session: Session, name: str):
#     return session.execute(
#         """
#         DELETE FROM scenario_results WHERE name=%s
#         """,
#         (name,),
#     )



from typing import Iterable

from cicadad.core.scenario import Scenario, filter_scenarios_by_tag, test_runner
from cicadad.protos import hub_pb2, hub_pb2_grpc


class HubServer(hub_pb2_grpc.HubServicer):
    def __init__(
        self,
        scenarios: Iterable[Scenario],
        image: str,
        network: str,
        datastore_address: str,
        container_service_address: str,
    ) -> None:
        self.scenarios = scenarios
        self.image = image
        self.network = network
        self.datastore_address = datastore_address
        self.container_service_address = container_service_address

    def Run(self, request, context):
        valid_scenarios = filter_scenarios_by_tag(self.scenarios, request.tags)

        yield hub_pb2.TestStatus(
            type="TEST_STARTED",
            message=f"Collected {len(valid_scenarios)} Scenario(s)",
        )

        for status in test_runner(
            valid_scenarios,
            self.image,
            self.network,
            self.datastore_address,
            self.container_service_address,
        ):
            yield hub_pb2.TestStatus(
                type=status.type,
                scenario=status.scenario,
                message=status.message,
                context=status.context,
            )

    def Healthcheck(self, request, context):
        return hub_pb2.HealthcheckReply(ready=True)


# FEATURE: run individual scenario



from google.protobuf import wrappers_pb2
import grpc  # type: ignore

from cicadad.core.containers import DockerServerArgs
from cicadad.protos import container_service_pb2, container_service_pb2_grpc
from cicadad.util.constants import DEFAULT_CONTAINER_SERVICE_ADDRESS


def start_container(
    args: DockerServerArgs,
    address: str = DEFAULT_CONTAINER_SERVICE_ADDRESS,
):
    with grpc.insecure_channel(address) as channel:
        stub = container_service_pb2_grpc.ContainerServiceStub(channel)
        request = container_service_pb2.StartContainerRequest(
            image=args.image,
            name=args.name,
            command=args.command,
            labels=args.labels,
            env=args.env,
            volumes=[]
            if args.volumes is None
            else [
                container_service_pb2.DockerVolume(
                    source=vol.source, destination=vol.destination
                )
                for vol in args.volumes
            ],
            hostPort=wrappers_pb2.Int32Value(value=args.host_port),
            containerPort=wrappers_pb2.Int32Value(value=args.container_port),
            network=wrappers_pb2.StringValue(value=args.network),
            createNetwork=wrappers_pb2.BoolValue(value=args.create_network),
        )

        stub.StartContainer(request)


def stop_container(
    container_id: str,
    address: str = DEFAULT_CONTAINER_SERVICE_ADDRESS,
):
    with grpc.insecure_channel(address) as channel:
        stub = container_service_pb2_grpc.ContainerServiceStub(channel)
        request = container_service_pb2.StopContainerRequest(containerID=container_id)

        stub.StopContainer(request)



from typing import Any
import base64
import json


def encode_context(context: Any) -> str:
    """Encode a context as a base64 string

    Args:
        context (any): Context object. Must be json serializable

    Returns:
        str: Encoded context base64 string
    """
    return base64.b64encode(json.dumps(context).encode("ascii")).decode("ascii")


def decode_context(encoded_context: str) -> Any:
    """Decode a base64 string to JSON deserialized context

    Args:
        encoded_context (str): Base64 context string

    Returns:
        any: Decoded context
    """
    decoded_context_bytes = base64.b64decode(
        encoded_context.encode("ascii"),
    )

    return json.loads(decoded_context_bytes.decode("ascii"))



from typing import Any, Callable, Union

from cicadad.core.scenario import (
    Scenario,
    LoadModelFn,
    UserLoopFn,
    ResultAggregatorFn,
    ResultVerifierFn,
    OutputTransformerFn,
)
from cicadad.core.engine import Engine


def scenario(engine: Engine, name: str = None):
    """Mark a function as a scenario

    Args:
        engine (Engine): Engine to attach scenario to
        name (str, optional): Name to give to scenario. Defaults to function name if None.
    """

    def wrapper(fn):
        if isinstance(fn, Scenario):
            raise TypeError("Function is already a Scenario")

        scenario = make_scenario(name=name or fn.__qualname__, fn=fn)

        engine.add_scenario(scenario)

        return scenario

    return wrapper


def user_loop(user_loop_fn: UserLoopFn):
    """Function to handle how the user function is run

    Args:
        user_loop_fn (UserLoopFn): User defined user loop function
    """

    def wrapper(fn):
        set_scenario_attribute(fn, "user_loop", user_loop_fn)

        return fn

    return wrapper


def load_model(load_model_fn: LoadModelFn):
    """Function to handle how scenario is run with regards to starting users and
    administering work

    Args:
        load_model_fn (LoadModelFn): User defined load model function
    """

    def wrapper(fn):
        set_scenario_attribute(fn, "load_model", load_model_fn)

        return fn

    return wrapper


def dependency(dep: Scenario):
    """Set a scenario as a dependency for this scenario to run

    Args:
        dep (Scenario): Scenario this function is dependent on being successful
    """

    def wrapper(fn):
        dependencies = get_scenario_attribute(fn, "dependencies")
        entry = [dep]

        if dependencies is None:
            set_scenario_attribute(fn, "dependencies", entry)
        else:
            set_scenario_attribute(fn, "dependencies", dependencies + entry)

        return fn

    return wrapper


def result_aggregator(aggregator_fn: ResultAggregatorFn):
    """Transform previous aggregate and list of results into an aggregated
    single result. Called by load model function

    Args:
        aggregator_fn (ResultAggregatorFn): Aggregator function
    """

    def wrapper(fn):
        set_scenario_attribute(fn, "result_aggregator", aggregator_fn)

        return fn

    return wrapper


def result_verifier(verifier_fn: ResultVerifierFn):
    """Create error messages for errors found in a list of results. Called by
    load model function

    Args:
        verifier_fn (ResultVerifierFn): Verifier function
    """

    def wrapper(fn):
        set_scenario_attribute(fn, "result_verifier", verifier_fn)

        return fn

    return wrapper


def output_transformer(transformer_fn: OutputTransformerFn):
    """Transform the aggregated result of the scenario after load model is
    called

    Args:
        transformer_fn (OutputTransformerFn): Transformer function
    """

    def wrapper(fn):
        set_scenario_attribute(fn, "output_transformer", transformer_fn)

        return fn

    return wrapper


def tag(tag: str):
    """Add a tag to a scenario. Tags allow for a test run to only cover certain
    scenarios.

    Args:
        tag (str): Tag name to add to scenario
    """

    def wrapper(fn):
        tags = get_scenario_attribute(fn, "tags")
        entry = [tag]

        if tags is None:
            set_scenario_attribute(fn, "tags", entry)
        else:
            set_scenario_attribute(fn, "tags", tags + entry)

        return fn

    return wrapper


def make_scenario(name: str, fn: Callable):
    if hasattr(fn, "__scenario_attribute__"):
        scenario_kwargs = fn.__scenario_attribute__  # type: ignore
        del fn.__scenario_attribute__  # type: ignore
    else:
        scenario_kwargs = {}

    return Scenario(name=name, fn=fn, **scenario_kwargs)


def get_scenario_attribute(obj: Union[Scenario, Callable], name: str):
    if hasattr(obj, name):
        return getattr(obj, name)

    if hasattr(obj, "__scenario_attribute__"):
        return obj.__scenario_attribute__.get(name)  # type: ignore

    return None


def set_scenario_attribute(obj: Union[Scenario, Callable], name: str, value: Any):
    if hasattr(obj, name):
        return setattr(obj, name, value)

    if hasattr(obj, "__scenario_attribute__"):
        obj.__scenario_attribute__[name] = value  # type: ignore
        return

    obj.__scenario_attribute__ = {name: value}  # type: ignore



import base64
import json

DEFAULT_DOCKER_NETWORK = "cicada-distributed-network"

DEFAULT_CONTEXT_STRING = base64.b64encode(json.dumps({}).encode("ascii")).decode(
    "ascii"
)

# METRICS_STREAM = "metrics"

LOG_FORMAT = "%(asctime)s | %(levelname)-8s | %(name)s:%(funcName)s -- %(message)s"
DATE_FORMAT = "%m-%d %H:%M"

DEFAULT_DATASTORE_ADDRESS = "cicada-distributed-datastore-client:8283"

DEFAULT_CONTAINER_SERVICE_ADDRESS = "cicada-distributed-container-service:8284"



from typing import Any, Callable
import time


def exponential_backoff(
    get_client_fn: Callable[[], Any],
    error_class: Any,
    tries: int,
    initial_wait: int = 2,
    multiplier: int = 2,
):
    """Get a client using an exponential backoff

    Args:
        get_client_fn (Callable): Function to get client.
        error_class (Any): Class to check exceptions against to determine if it should retry.
        tries (int): Number of attempts allowed to get client.
        initial_wait (int, optional): Time in seconds for first wait after a failure. Defaults to 2.
        multiplier (int, optional): Amount to multiply wait time by after each failure. Defaults to 2.

    Raises:
        e: Exception if not an instance of error_class.
        RuntimeError: Exhausted retries.

    Returns:
        Any: Client if successfully configured
    """
    remaining_tries = tries
    wait = initial_wait

    while remaining_tries > 0:
        try:
            return get_client_fn()
        except Exception as e:
            if isinstance(e, error_class):
                time.sleep(wait)

                remaining_tries -= 1
                wait *= multiplier
            else:
                raise e

    raise RuntimeError(f"Timed out waiting for client: {get_client_fn}")



from typing import Dict
from concurrent import futures

import click
import grpc  # type: ignore

from cicadad.core.scenario import Scenario, scenario_runner, user_runner
from cicadad.util.constants import (
    DEFAULT_CONTAINER_SERVICE_ADDRESS,
    DEFAULT_DATASTORE_ADDRESS,
    DEFAULT_CONTEXT_STRING,
    DEFAULT_DOCKER_NETWORK,
)
from cicadad.util.context import decode_context
from cicadad.server.hub import HubServer
from cicadad.protos import hub_pb2_grpc


class Engine:
    def __init__(self) -> None:
        """Entrypoint for Cicada tests. Links tests to Cicada infrastructure"""
        self.scenarios: Dict[str, Scenario] = {}

    def add_scenario(self, scenario: Scenario):
        """Add a scenario to the engine

        Args:
            scenario (Scenario): Scenario being added to engine
        """
        self.scenarios[scenario.name] = scenario

    def start(self):
        """Called internally when test container is started to parse args"""
        # read sys.argv and start scenario or user (Already in container)
        engine_cli(obj=self)

    def run_test(
        self,
        image: str,
        network: str,
        datastore_address: str,
        container_service_address: str,
    ):
        """Startup function when test container is created. Starts hub server.

        Args:
            image (str): Image containing test code. Passed to scenarios
            network (str): Network to start scenarios in
            datastore_address (str): Address of datastore client to receive scenario results
            container_service_address (str): Address of container service to start scenarios
        """
        server = grpc.server(futures.ThreadPoolExecutor())

        hub_pb2_grpc.add_HubServicer_to_server(
            HubServer(
                scenarios=self.scenarios.values(),
                image=image,
                network=network,
                datastore_address=datastore_address,
                container_service_address=container_service_address,
            ),
            server,
        )

        server.add_insecure_port("[::]:50051")
        server.start()
        server.wait_for_termination()

    def run_scenario(
        self,
        scenario_name: str,
        scenario_id: str,
        image: str,
        network: str,
        datastore_address: str,
        container_service_address: str,
        encoded_context: str,
    ):
        """Startup function when scenario is started. Begins running scenario's
        load function

        Args:
            scenario_name (str): Name of scenario being run
            scenario_id (str): ID of scenario run
            image (str): Image containing test code
            network (str): Network to start users in
            datastore_address (str): Address of datastore client to save and receive results
            container_service_address (str): Address of container service to start and stop users
            encoded_context (str): Context from test containing previous results
        """
        scenario = self.scenarios[scenario_name]
        context = decode_context(encoded_context)

        scenario_runner(
            scenario,
            image,
            network,
            scenario_id,
            datastore_address,
            container_service_address,
            context,
        )

    def run_user(
        self,
        scenario_name: str,
        user_id: str,
        datastore_address: str,
        encoded_context: str,
    ):
        """Startup function when user is started. Runs scenario user loop.

        Args:
            scenario_name (str): Name of scenario being run
            user_id (str): Unique ID of user assigned by scenario
            datastore_address (str): Address of datastore client to receive work and save results
            encoded_context (str): Context from test containing previous results
        """
        scenario = self.scenarios[scenario_name]
        context = decode_context(encoded_context)

        user_runner(
            scenario,
            user_id,
            datastore_address,
            context,
        )


@click.group()
@click.pass_context
def engine_cli(ctx):
    ctx.ensure_object(Engine)


@engine_cli.command()
@click.pass_context
@click.option("--image", type=str, required=True)
@click.option("--network", type=str, default=DEFAULT_DOCKER_NETWORK)
@click.option("--datastore-address", type=str, default=DEFAULT_DATASTORE_ADDRESS)
@click.option(
    "--container-service-address", type=str, default=DEFAULT_CONTAINER_SERVICE_ADDRESS
)
def run_test(ctx, image, network, datastore_address, container_service_address):
    engine: Engine = ctx.obj

    engine.run_test(
        image,
        network,
        datastore_address,
        container_service_address,
    )


@engine_cli.command()
@click.pass_context
@click.option("--name", type=str, required=True)
@click.option("--scenario-id", type=str, required=True)
@click.option("--image", type=str, required=True)
@click.option("--network", type=str, default=DEFAULT_DOCKER_NETWORK)
@click.option("--datastore-address", type=str, default=DEFAULT_DATASTORE_ADDRESS)
@click.option(
    "--container-service-address", type=str, default=DEFAULT_CONTAINER_SERVICE_ADDRESS
)
@click.option(
    "--encoded-context",
    type=str,
    default=DEFAULT_CONTEXT_STRING,
)
def run_scenario(
    ctx,
    name,
    scenario_id,
    image,
    network,
    datastore_address,
    container_service_address,
    encoded_context,
):
    engine: Engine = ctx.obj

    engine.run_scenario(
        name,
        scenario_id,
        image,
        network,
        datastore_address,
        container_service_address,
        encoded_context,
    )


@engine_cli.command()
@click.pass_context
@click.option("--name", type=str, required=True)
@click.option("--user-id", type=str, required=True)
@click.option("--datastore-address", type=str, default=DEFAULT_DATASTORE_ADDRESS)
@click.option(
    "--encoded-context",
    type=str,
    default=DEFAULT_CONTEXT_STRING,
)
def run_user(
    ctx,
    name,
    user_id,
    datastore_address,
    encoded_context,
):
    engine: Engine = ctx.obj

    engine.run_user(
        name,
        user_id,
        datastore_address,
        encoded_context,
    )



from datetime import datetime
from unittest.mock import Mock, patch, call
from pytest import raises

from cicadad.core import scenario as scenario_module
from cicadad.services import datastore


@patch("cicadad.services.datastore.get_work")
def test_has_work(get_work_mock):
    user_id = "abc"
    address = "some address"
    s = Mock()
    get_work_mock.return_value = 1

    uc = scenario_module.UserCommands(s, user_id, address)

    assert uc.has_work()
    assert uc.available_work == 0

    get_work_mock.assert_called_with(user_id, address)


@patch("cicadad.services.datastore.get_work")
def test_has_more_work(get_work_mock):
    user_id = "abc"
    address = "some address"
    s = Mock()
    get_work_mock.return_value = 2

    uc = scenario_module.UserCommands(s, user_id, address)

    assert uc.has_work(500)
    assert uc.has_work(500)
    assert uc.available_work == 0

    get_work_mock.assert_called_once_with(user_id, address)


@patch("cicadad.services.datastore.get_work")
def test_has_no_work(get_work_mock):
    user_id = "abc"
    address = "some address"
    s = Mock()
    get_work_mock.return_value = 0

    uc = scenario_module.UserCommands(s, user_id, address)

    assert not uc.has_work()
    assert get_work_mock.call_count == 2

    get_work_mock.assert_called_with(user_id, address)


def test_run_logs():
    def test_fn():
        print("foo")

    user_id = "abc"
    address = "some address"
    s = Mock()

    s.fn = test_fn

    uc = scenario_module.UserCommands(s, user_id, address)

    output, exception, logs = uc.run()

    assert output is None
    assert exception is None
    assert logs == "foo\n"


def test_run_output():
    def test_fn():
        return 42

    user_id = "abc"
    address = "some address"
    s = Mock()

    s.fn = test_fn

    uc = scenario_module.UserCommands(s, user_id, address)

    output, exception, logs = uc.run()

    assert output == 42
    assert exception is None
    assert logs == ""


def test_run_exception():
    def test_fn():
        raise ValueError("some error")

    user_id = "abc"
    address = "some address"
    s = Mock()

    s.fn = test_fn

    uc = scenario_module.UserCommands(s, user_id, address)

    output, exception, logs = uc.run(log_traceback=False)

    assert output is None
    assert isinstance(exception, ValueError)
    assert str(exception) == "some error"
    assert logs == ""


def test_scale_users_up():
    s = Mock()
    image = "foo"
    network = "bar"
    sid = "abc"
    datastore_addr = "fizz"
    container_service_addr = "buzz"
    ctx = {}

    sc = scenario_module.ScenarioCommands(
        s,
        image,
        network,
        sid,
        datastore_addr,
        container_service_addr,
        ctx,
    )

    sc.start_users = Mock()
    sc.stop_users = Mock()

    sc.scale_users(10)

    sc.start_users.assert_called_once_with(10)


def test_scale_users_down():
    s = Mock()
    image = "foo"
    network = "bar"
    sid = "abc"
    datastore_addr = "fizz"
    container_service_addr = "buzz"
    ctx = {}

    sc = scenario_module.ScenarioCommands(
        s,
        image,
        network,
        sid,
        datastore_addr,
        container_service_addr,
        ctx,
    )

    sc.start_users = Mock()
    sc.stop_users = Mock()

    sc.num_users = 20

    sc.scale_users(10)

    sc.stop_users.assert_called_once_with(10)


@patch("cicadad.services.container_service.start_container")
def test_start_users(start_container_mock):
    s = Mock()
    image = "foo"
    network = "bar"
    sid = "abc"
    datastore_addr = "fizz"
    container_service_addr = "buzz"
    ctx = {}

    s.name = "s"

    sc = scenario_module.ScenarioCommands(
        s,
        image,
        network,
        sid,
        datastore_addr,
        container_service_addr,
        ctx,
    )

    sc.buffered_work = 10
    sc.add_work = Mock()

    sc.start_users(5)

    assert start_container_mock.call_count == 5

    sc.add_work.assert_called_once_with(10)
    assert sc.buffered_work == 0


def test_start_users_negative():
    s = Mock()
    image = "foo"
    network = "bar"
    sid = "abc"
    datastore_addr = "fizz"
    container_service_addr = "buzz"
    ctx = {}

    sc = scenario_module.ScenarioCommands(
        s,
        image,
        network,
        sid,
        datastore_addr,
        container_service_addr,
        ctx,
    )

    with raises(ValueError, match="Must supply a positive number of users to start"):
        sc.start_users(-1)


@patch("cicadad.services.container_service.stop_container")
def test_stop_users(stop_container_mock):
    s = Mock()
    image = "foo"
    network = "bar"
    sid = "abc"
    datastore_addr = "fizz"
    container_service_addr = "buzz"
    ctx = {}

    sc = scenario_module.ScenarioCommands(
        s,
        image,
        network,
        sid,
        datastore_addr,
        container_service_addr,
        ctx,
    )

    stop_container_mock.return_value = None

    sc.num_users = 4
    sc.user_ids = ["1", "2", "3", "4"]

    sc.stop_users(3)

    assert sc.num_users == 1
    assert sc.user_ids == ["4"]


def test_stop_users_too_many():
    s = Mock()
    image = "foo"
    network = "bar"
    sid = "abc"
    datastore_addr = "fizz"
    container_service_addr = "buzz"
    ctx = {}

    sc = scenario_module.ScenarioCommands(
        s,
        image,
        network,
        sid,
        datastore_addr,
        container_service_addr,
        ctx,
    )

    with raises(ValueError, match="Scenario currently has less than 3 users"):
        sc.stop_users(3)


def test_stop_users_negative():
    s = Mock()
    image = "foo"
    network = "bar"
    sid = "abc"
    datastore_addr = "fizz"
    container_service_addr = "buzz"
    ctx = {}

    sc = scenario_module.ScenarioCommands(
        s,
        image,
        network,
        sid,
        datastore_addr,
        container_service_addr,
        ctx,
    )

    with raises(ValueError, match="Must supply a positive number of users to stop"):
        sc.stop_users(-1)


@patch("cicadad.services.datastore.distribute_work")
def test_add_work(distribute_work_mock):
    s = Mock()
    image = "foo"
    network = "bar"
    sid = "abc"
    datastore_addr = "fizz"
    container_service_addr = "buzz"
    ctx = {}

    sc = scenario_module.ScenarioCommands(
        s,
        image,
        network,
        sid,
        datastore_addr,
        container_service_addr,
        ctx,
    )

    distribute_work_mock.return_value = None

    sc.num_users = 3
    sc.user_ids = ["1", "2", "3"]

    sc.add_work(11)

    assert distribute_work_mock.call_count == 1


def test_has_work_buffered():
    s = Mock()
    image = "foo"
    network = "bar"
    sid = "abc"
    datastore_addr = "fizz"
    container_service_addr = "buzz"
    ctx = {}

    sc = scenario_module.ScenarioCommands(
        s,
        image,
        network,
        sid,
        datastore_addr,
        container_service_addr,
        ctx,
    )

    sc.add_work(10)

    assert sc.buffered_work == 10


def test_aggregate_results():
    s = Mock()
    image = "foo"
    network = "bar"
    sid = "abc"
    datastore_addr = "fizz"
    container_service_addr = "buzz"
    ctx = {}

    sc = scenario_module.ScenarioCommands(
        s,
        image,
        network,
        sid,
        datastore_addr,
        container_service_addr,
        ctx,
    )

    def aggregator_fn(previous, latest_results):
        if previous is None:
            p = 0
        else:
            p = previous

        return p + sum(latest_results)

    s.result_aggregator = aggregator_fn

    assert sc.aggregate_results([1]) == 1
    assert sc.aggregate_results([2]) == 3
    assert sc.aggregate_results([3]) == 6

    assert sc.aggregated_results == 6


def test_aggregate_results_default():
    s = Mock()
    image = "foo"
    network = "bar"
    sid = "abc"
    datastore_addr = "fizz"
    container_service_addr = "buzz"
    ctx = {}

    sc = scenario_module.ScenarioCommands(
        s,
        image,
        network,
        sid,
        datastore_addr,
        container_service_addr,
        ctx,
    )

    s.result_aggregator = None

    assert sc.aggregate_results([datastore.Result(output=1)]) == 1
    assert sc.aggregate_results([datastore.Result(output=2)]) == 2
    assert sc.aggregate_results([datastore.Result(output=3)]) == 3

    assert sc.aggregated_results == 3


def test_verify_results():
    s = Mock()
    image = "foo"
    network = "bar"
    sid = "abc"
    datastore_addr = "fizz"
    container_service_addr = "buzz"
    ctx = {}

    sc = scenario_module.ScenarioCommands(
        s,
        image,
        network,
        sid,
        datastore_addr,
        container_service_addr,
        ctx,
    )

    def result_verifier(results):
        return ["error" for r in results if not r]

    s.result_verifier = result_verifier

    assert sc.verify_results([False, True]) == ["error"]


def test_filter_scenarios():
    s1 = Mock()
    s2 = Mock()

    s1.tags = ["foo", "bar"]
    s2.tags = ["fizz", "buzz"]

    assert scenario_module.filter_scenarios_by_tag([s1, s2], ["fizz", "bizz"]) == [s2]


def test_filter_scenarios_empty():
    s1 = Mock()
    s2 = Mock()

    assert scenario_module.filter_scenarios_by_tag([s1, s2], []) == [s1, s2]


@patch("cicadad.services.datastore.move_scenario_result")
@patch("cicadad.services.container_service.start_container")
def test_test_runner(start_container_mock, move_scenario_event_mock):
    s1 = Mock()
    s2 = Mock()
    s3 = Mock()

    s1.name = "s1"
    s2.name = "s2"
    s3.name = "s3"

    s1.dependencies = []
    s2.dependencies = [s1]
    s3.dependencies = [s2]

    ss = [s1, s2, s3]
    img = "foo"
    n = "bar"
    datastore_addr = "fizz"
    container_service_addr = "buzz"

    r1 = {
        "output": "42",
        "exception": None,
        "logs": "",
        "timestamp": str(datetime.now()),
        "time_taken": 3,
    }

    r2 = {
        "output": None,
        "exception": "some error",
        "logs": "",
        "timestamp": str(datetime.now()),
        "time_taken": 3,
    }

    move_scenario_event_mock.side_effect = [r1, r2]

    assert (
        len(
            list(
                scenario_module.test_runner(
                    ss, img, n, datastore_addr, container_service_addr
                )
            )
        )
        == 5
    )
    assert start_container_mock.call_count == 2


@patch("cicadad.services.datastore.set_scenario_result")
def test_run_scenario(set_scenario_result_mock):
    s = Mock()
    image = "foo"
    network = "bar"
    sid = "abc"
    datastore_addr = "fizz"
    container_service_addr = "buzz"
    ctx = {}

    def load_model(sc, c):
        sc.aggregated_results = 42
        sc.stop_users = Mock()

    s.load_model = load_model
    s.output_transformer = None

    scenario_module.scenario_runner(
        s,
        image,
        network,
        sid,
        datastore_addr,
        container_service_addr,
        ctx,
    )

    # Get kwargs of first call
    assert set_scenario_result_mock.mock_calls[0][2]["output"] == 42


@patch("cicadad.services.datastore.set_scenario_result")
def test_run_scenario_result_transformer(set_scenario_result_mock):
    s = Mock()
    image = "foo"
    network = "bar"
    sid = "abc"
    datastore_addr = "fizz"
    container_service_addr = "buzz"
    ctx = {}

    def load_model(sc, c):
        sc.aggregated_results = 42
        sc.stop_users = Mock()

    def double_result(ar):
        return ar * 2

    s.load_model = load_model
    s.output_transformer = double_result

    scenario_module.scenario_runner(
        s,
        image,
        network,
        sid,
        datastore_addr,
        container_service_addr,
        ctx,
    )

    # Get kwargs of first call
    assert set_scenario_result_mock.mock_calls[0][2]["output"] == 84


@patch("cicadad.services.datastore.set_scenario_result")
def test_run_scenario_exception(set_scenario_result_mock):
    s = Mock()
    image = "foo"
    network = "bar"
    sid = "abc"
    datastore_addr = "fizz"
    container_service_addr = "buzz"
    ctx = {}

    def load_model(sc, c):
        sc.aggregated_results = None
        sc.errors = ["some error"]
        sc.stop_users = Mock()

    s.load_model = load_model
    s.output_transformer = None
    s.name = "s"

    scenario_module.scenario_runner(
        s,
        image,
        network,
        sid,
        datastore_addr,
        container_service_addr,
        ctx,
    )

    # Get first call, then args of call, then 4th arg of call
    assert (
        str(set_scenario_result_mock.mock_calls[0][2]["exception"])
        == "1 error(s) were raised in scenario s:\nsome error"
    )

    assert (
        type(set_scenario_result_mock.mock_calls[0][2]["exception"]) == AssertionError
    )


def test_while_has_work():
    closure = scenario_module.while_has_work(500)

    uc = Mock()
    ctx = {}

    uc.is_up.side_effect = [True, True, False]
    uc.has_work.side_effect = [True, False]
    uc.run.return_value = 42, None, ""

    closure(uc, ctx)

    uc.has_work.assert_called_with(500)
    assert uc.report_result.call_count == 1


def test_while_alive():
    closure = scenario_module.while_alive()

    uc = Mock()
    ctx = {}

    uc.is_up.side_effect = [True, True, False]
    uc.run.return_value = 42, None, ""

    closure(uc, ctx)

    assert uc.report_result.call_count == 2


def test_iterations_per_second_limited():
    closure = scenario_module.iterations_per_second_limited(3)

    uc = Mock()
    ctx = {}

    uc.is_up.side_effect = [True, True, True, True, False]
    uc.run.return_value = 42, None, ""

    closure(uc, ctx)

    assert uc.report_result.call_count == 3


@patch("cicadad.core.scenario.time")
def test_n_iterations(time_mock):
    closure = scenario_module.n_iterations(3, 2)

    sc = Mock()
    ctx = {}

    sc.get_latest_results.side_effect = [[1], [2, 3]]

    closure(sc, ctx)

    assert sc.scale_users.call_count == 2
    assert sc.scale_users.mock_calls[0] == call(2)
    sc.add_work.assert_called_with(3)


def test_n_iterations_timeout():
    closure = scenario_module.n_iterations(3, 2, wait_period=1, timeout=1)

    sc = Mock()
    ctx = {}

    sc.get_latest_results.side_effect = [[], []]

    with raises(AssertionError, match="Timed out waiting for results"):
        closure(sc, ctx)


def test_n_seconds():
    closure = scenario_module.n_seconds(2, 2)

    sc = Mock()
    ctx = {}

    sc.get_latest_results.side_effect = [[1], [2, 3]]

    closure(sc, ctx)

    assert sc.scale_users.call_count == 2
    assert sc.scale_users.mock_calls[0] == call(2)


def test_n_users_ramping_add_users():
    closure = scenario_module.n_users_ramping(6, 5)

    class ScenarioCommandsMock:
        def __init__(self):
            self.num_users = 0
            self.start_users_calls = 0

        def start_users(self, n):
            self.num_users += n
            self.start_users_calls += 1

        def scale_users(self, n):
            self.num_users = n

    sc = ScenarioCommandsMock()
    ctx = {}

    sc.get_latest_results = Mock()
    sc.aggregate_results = Mock()
    sc.verify_results = Mock()

    closure(sc, ctx)

    assert sc.num_users == 5
    assert sc.start_users_calls == 5


def test_n_users_ramping_stop_users():
    closure = scenario_module.n_users_ramping(5, 6)

    class ScenarioCommandsMock:
        def __init__(self):
            self.num_users = 10
            self.stop_users_calls = 0

        def stop_users(self, n):
            self.num_users -= n
            self.stop_users_calls += 1

        def scale_users(self, n):
            self.num_users = n

    sc = ScenarioCommandsMock()
    ctx = {}

    sc.get_latest_results = Mock()
    sc.aggregate_results = Mock()
    sc.verify_results = Mock()

    closure(sc, ctx)

    assert sc.num_users == 6
    assert sc.stop_users_calls == 4


def test_load_stages():
    s1 = scenario_module.n_seconds(2, 2, skip_scaledown=True)
    s2 = scenario_module.n_iterations(2, 4, skip_scaledown=True)

    closure = scenario_module.load_stages(s1, s2)

    sc = Mock()
    ctx = {}

    sc.get_latest_results.side_effect = [[1], [2, 3], [4], [5, 6]]

    closure(sc, ctx)

    assert sc.scale_users.call_count == 3
    assert sc.scale_users.mock_calls[0] == call(2)
    assert sc.scale_users.mock_calls[1] == call(4)
    assert sc.scale_users.mock_calls[2] == call(0)



from unittest.mock import Mock

from cicadad.core import decorators, scenario


def test_scenario():
    e = Mock()

    @decorators.scenario(e, "my-test")
    def test_fn():
        return 42

    assert isinstance(test_fn, scenario.Scenario)
    assert test_fn.name == "my-test"
    assert test_fn.user_loop is not None
    assert test_fn.load_model is not None
    assert test_fn.result_verifier is not None
    assert test_fn.dependencies == []
    assert test_fn.tags == []


def test_user_loop():
    e = Mock()
    ul = Mock()

    @decorators.scenario(e)
    @decorators.user_loop(ul)
    def test_fn():
        return 42

    assert test_fn.user_loop == ul


def test_load_model():
    e = Mock()
    lm = Mock()

    @decorators.scenario(e)
    @decorators.load_model(lm)
    def test_fn():
        return 42

    assert test_fn.load_model == lm


def test_dependency():
    e = Mock()

    @decorators.scenario(e)
    def test_fn_1():
        return 42

    @decorators.scenario(e)
    @decorators.dependency(test_fn_1)
    def test_fn_2():
        return 42

    assert test_fn_2.dependencies == [test_fn_1]


def test_result_aggregator():
    e = Mock()
    ra = Mock()

    @decorators.scenario(e)
    @decorators.result_aggregator(ra)
    def test_fn():
        return 42

    assert test_fn.result_aggregator == ra


def test_result_verifier():
    e = Mock()
    rv = Mock()

    @decorators.scenario(e)
    @decorators.result_verifier(rv)
    def test_fn():
        return 42

    assert test_fn.result_verifier == rv


def test_result_verifier_none():
    e = Mock()

    @decorators.scenario(e)
    @decorators.result_verifier(None)
    def test_fn():
        return 42

    assert test_fn.result_verifier is None


def test_output_transformer():
    e = Mock()
    ot = Mock()

    @decorators.scenario(e)
    @decorators.output_transformer(ot)
    def test_fn():
        return 42

    assert test_fn.output_transformer == ot


def test_tags():
    e = Mock()

    @decorators.scenario(e)
    @decorators.tag("foo")
    def test_fn():
        return 42

    assert test_fn.tags == ["foo"]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cicadad/protos/hub.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='cicadad/protos/hub.proto',
  package='cicada_distributed',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x18\x63icadad/protos/hub.proto\x12\x12\x63icada_distributed\x1a\x1bgoogle/protobuf/empty.proto\"\x1e\n\x0eRunTestRequest\x12\x0c\n\x04tags\x18\x01 \x03(\t\"N\n\nTestStatus\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x10\n\x08scenario\x18\x02 \x01(\t\x12\x0f\n\x07message\x18\x03 \x01(\t\x12\x0f\n\x07\x63ontext\x18\x04 \x01(\t\"!\n\x10HealthcheckReply\x12\r\n\x05ready\x18\x01 \x01(\x08\x32\x9f\x01\n\x03Hub\x12K\n\x03Run\x12\".cicada_distributed.RunTestRequest\x1a\x1e.cicada_distributed.TestStatus0\x01\x12K\n\x0bHealthcheck\x12\x16.google.protobuf.Empty\x1a$.cicada_distributed.HealthcheckReplyb\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_empty__pb2.DESCRIPTOR,])




_RUNTESTREQUEST = _descriptor.Descriptor(
  name='RunTestRequest',
  full_name='cicada_distributed.RunTestRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='tags', full_name='cicada_distributed.RunTestRequest.tags', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=77,
  serialized_end=107,
)


_TESTSTATUS = _descriptor.Descriptor(
  name='TestStatus',
  full_name='cicada_distributed.TestStatus',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='cicada_distributed.TestStatus.type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='scenario', full_name='cicada_distributed.TestStatus.scenario', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message', full_name='cicada_distributed.TestStatus.message', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='context', full_name='cicada_distributed.TestStatus.context', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=109,
  serialized_end=187,
)


_HEALTHCHECKREPLY = _descriptor.Descriptor(
  name='HealthcheckReply',
  full_name='cicada_distributed.HealthcheckReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='ready', full_name='cicada_distributed.HealthcheckReply.ready', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=189,
  serialized_end=222,
)

DESCRIPTOR.message_types_by_name['RunTestRequest'] = _RUNTESTREQUEST
DESCRIPTOR.message_types_by_name['TestStatus'] = _TESTSTATUS
DESCRIPTOR.message_types_by_name['HealthcheckReply'] = _HEALTHCHECKREPLY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

RunTestRequest = _reflection.GeneratedProtocolMessageType('RunTestRequest', (_message.Message,), {
  'DESCRIPTOR' : _RUNTESTREQUEST,
  '__module__' : 'cicadad.protos.hub_pb2'
  # @@protoc_insertion_point(class_scope:cicada_distributed.RunTestRequest)
  })
_sym_db.RegisterMessage(RunTestRequest)

TestStatus = _reflection.GeneratedProtocolMessageType('TestStatus', (_message.Message,), {
  'DESCRIPTOR' : _TESTSTATUS,
  '__module__' : 'cicadad.protos.hub_pb2'
  # @@protoc_insertion_point(class_scope:cicada_distributed.TestStatus)
  })
_sym_db.RegisterMessage(TestStatus)

HealthcheckReply = _reflection.GeneratedProtocolMessageType('HealthcheckReply', (_message.Message,), {
  'DESCRIPTOR' : _HEALTHCHECKREPLY,
  '__module__' : 'cicadad.protos.hub_pb2'
  # @@protoc_insertion_point(class_scope:cicada_distributed.HealthcheckReply)
  })
_sym_db.RegisterMessage(HealthcheckReply)



_HUB = _descriptor.ServiceDescriptor(
  name='Hub',
  full_name='cicada_distributed.Hub',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=225,
  serialized_end=384,
  methods=[
  _descriptor.MethodDescriptor(
    name='Run',
    full_name='cicada_distributed.Hub.Run',
    index=0,
    containing_service=None,
    input_type=_RUNTESTREQUEST,
    output_type=_TESTSTATUS,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Healthcheck',
    full_name='cicada_distributed.Hub.Healthcheck',
    index=1,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=_HEALTHCHECKREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_HUB)

DESCRIPTOR.services_by_name['Hub'] = _HUB

# @@protoc_insertion_point(module_scope)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from cicadad.protos import datastore_pb2 as cicadad_dot_protos_dot_datastore__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


class DatastoreStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.AddUserResult = channel.unary_unary(
                '/datastore.Datastore/AddUserResult',
                request_serializer=cicadad_dot_protos_dot_datastore__pb2.AddUserResultRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.SetScenarioResult = channel.unary_unary(
                '/datastore.Datastore/SetScenarioResult',
                request_serializer=cicadad_dot_protos_dot_datastore__pb2.SetScenarioResultRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.MoveUserResults = channel.unary_unary(
                '/datastore.Datastore/MoveUserResults',
                request_serializer=cicadad_dot_protos_dot_datastore__pb2.MoveUserResultsRequest.SerializeToString,
                response_deserializer=cicadad_dot_protos_dot_datastore__pb2.MoveUserResultsResponse.FromString,
                )
        self.MoveScenarioResult = channel.unary_unary(
                '/datastore.Datastore/MoveScenarioResult',
                request_serializer=cicadad_dot_protos_dot_datastore__pb2.MoveScenarioResultRequest.SerializeToString,
                response_deserializer=cicadad_dot_protos_dot_datastore__pb2.MoveScenarioResultResponse.FromString,
                )
        self.DistributeWork = channel.unary_unary(
                '/datastore.Datastore/DistributeWork',
                request_serializer=cicadad_dot_protos_dot_datastore__pb2.DistributeWorkRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.GetUserWork = channel.unary_unary(
                '/datastore.Datastore/GetUserWork',
                request_serializer=cicadad_dot_protos_dot_datastore__pb2.GetUserWorkRequest.SerializeToString,
                response_deserializer=cicadad_dot_protos_dot_datastore__pb2.GetUserWorkResponse.FromString,
                )


class DatastoreServicer(object):
    """Missing associated documentation comment in .proto file."""

    def AddUserResult(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetScenarioResult(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def MoveUserResults(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def MoveScenarioResult(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DistributeWork(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetUserWork(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_DatastoreServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'AddUserResult': grpc.unary_unary_rpc_method_handler(
                    servicer.AddUserResult,
                    request_deserializer=cicadad_dot_protos_dot_datastore__pb2.AddUserResultRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'SetScenarioResult': grpc.unary_unary_rpc_method_handler(
                    servicer.SetScenarioResult,
                    request_deserializer=cicadad_dot_protos_dot_datastore__pb2.SetScenarioResultRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'MoveUserResults': grpc.unary_unary_rpc_method_handler(
                    servicer.MoveUserResults,
                    request_deserializer=cicadad_dot_protos_dot_datastore__pb2.MoveUserResultsRequest.FromString,
                    response_serializer=cicadad_dot_protos_dot_datastore__pb2.MoveUserResultsResponse.SerializeToString,
            ),
            'MoveScenarioResult': grpc.unary_unary_rpc_method_handler(
                    servicer.MoveScenarioResult,
                    request_deserializer=cicadad_dot_protos_dot_datastore__pb2.MoveScenarioResultRequest.FromString,
                    response_serializer=cicadad_dot_protos_dot_datastore__pb2.MoveScenarioResultResponse.SerializeToString,
            ),
            'DistributeWork': grpc.unary_unary_rpc_method_handler(
                    servicer.DistributeWork,
                    request_deserializer=cicadad_dot_protos_dot_datastore__pb2.DistributeWorkRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'GetUserWork': grpc.unary_unary_rpc_method_handler(
                    servicer.GetUserWork,
                    request_deserializer=cicadad_dot_protos_dot_datastore__pb2.GetUserWorkRequest.FromString,
                    response_serializer=cicadad_dot_protos_dot_datastore__pb2.GetUserWorkResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'datastore.Datastore', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Datastore(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def AddUserResult(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/datastore.Datastore/AddUserResult',
            cicadad_dot_protos_dot_datastore__pb2.AddUserResultRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetScenarioResult(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/datastore.Datastore/SetScenarioResult',
            cicadad_dot_protos_dot_datastore__pb2.SetScenarioResultRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def MoveUserResults(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/datastore.Datastore/MoveUserResults',
            cicadad_dot_protos_dot_datastore__pb2.MoveUserResultsRequest.SerializeToString,
            cicadad_dot_protos_dot_datastore__pb2.MoveUserResultsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def MoveScenarioResult(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/datastore.Datastore/MoveScenarioResult',
            cicadad_dot_protos_dot_datastore__pb2.MoveScenarioResultRequest.SerializeToString,
            cicadad_dot_protos_dot_datastore__pb2.MoveScenarioResultResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DistributeWork(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/datastore.Datastore/DistributeWork',
            cicadad_dot_protos_dot_datastore__pb2.DistributeWorkRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetUserWork(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/datastore.Datastore/GetUserWork',
            cicadad_dot_protos_dot_datastore__pb2.GetUserWorkRequest.SerializeToString,
            cicadad_dot_protos_dot_datastore__pb2.GetUserWorkResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from cicadad.protos import hub_pb2 as cicadad_dot_protos_dot_hub__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


class HubStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Run = channel.unary_stream(
                '/cicada_distributed.Hub/Run',
                request_serializer=cicadad_dot_protos_dot_hub__pb2.RunTestRequest.SerializeToString,
                response_deserializer=cicadad_dot_protos_dot_hub__pb2.TestStatus.FromString,
                )
        self.Healthcheck = channel.unary_unary(
                '/cicada_distributed.Hub/Healthcheck',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=cicadad_dot_protos_dot_hub__pb2.HealthcheckReply.FromString,
                )


class HubServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Run(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Healthcheck(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_HubServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Run': grpc.unary_stream_rpc_method_handler(
                    servicer.Run,
                    request_deserializer=cicadad_dot_protos_dot_hub__pb2.RunTestRequest.FromString,
                    response_serializer=cicadad_dot_protos_dot_hub__pb2.TestStatus.SerializeToString,
            ),
            'Healthcheck': grpc.unary_unary_rpc_method_handler(
                    servicer.Healthcheck,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=cicadad_dot_protos_dot_hub__pb2.HealthcheckReply.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'cicada_distributed.Hub', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Hub(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Run(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/cicada_distributed.Hub/Run',
            cicadad_dot_protos_dot_hub__pb2.RunTestRequest.SerializeToString,
            cicadad_dot_protos_dot_hub__pb2.TestStatus.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Healthcheck(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cicada_distributed.Hub/Healthcheck',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            cicadad_dot_protos_dot_hub__pb2.HealthcheckReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



from cicadad.core.decorators import scenario
from cicadad.core.engine import Engine

engine = Engine()


@scenario(engine)
def my_first_test(context):
    # Results of previously run scenarios get passed in as context

    # Write the body of your test here
    assert 2 + 2 == 4

    # Anything returned gets saved as a user output
    return "Passed!"


if __name__ == "__main__":
    engine.start()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cicadad/protos/datastore.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='cicadad/protos/datastore.proto',
  package='datastore',
  syntax='proto3',
  serialized_options=b'Z-github.com/cicadatesting/datastore-client/api',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1e\x63icadad/protos/datastore.proto\x12\tdatastore\x1a\x1bgoogle/protobuf/empty.proto\x1a\x1egoogle/protobuf/wrappers.proto\"6\n\x14\x41\x64\x64UserResultRequest\x12\x0e\n\x06userID\x18\x01 \x01(\t\x12\x0e\n\x06result\x18\x02 \x01(\x0c\"\xae\x01\n\x18SetScenarioResultRequest\x12\x12\n\nscenarioID\x18\x01 \x01(\t\x12,\n\x06output\x18\x02 \x01(\x0b\x32\x1c.google.protobuf.StringValue\x12/\n\texception\x18\x03 \x01(\x0b\x32\x1c.google.protobuf.StringValue\x12\x0c\n\x04logs\x18\x04 \x01(\t\x12\x11\n\ttimeTaken\x18\x05 \x01(\x01\")\n\x16MoveUserResultsRequest\x12\x0f\n\x07userIDs\x18\x01 \x03(\t\"*\n\x17MoveUserResultsResponse\x12\x0f\n\x07results\x18\x01 \x03(\x0c\"/\n\x19MoveScenarioResultRequest\x12\x12\n\nscenarioID\x18\x01 \x01(\t\"\xbb\x01\n\x1aMoveScenarioResultResponse\x12\n\n\x02id\x18\x01 \x01(\t\x12,\n\x06output\x18\x02 \x01(\x0b\x32\x1c.google.protobuf.StringValue\x12/\n\texception\x18\x03 \x01(\x0b\x32\x1c.google.protobuf.StringValue\x12\x0c\n\x04logs\x18\x04 \x01(\t\x12\x11\n\ttimestamp\x18\x05 \x01(\t\x12\x11\n\ttimeTaken\x18\x06 \x01(\x01\"6\n\x15\x44istributeWorkRequest\x12\x0c\n\x04work\x18\x01 \x01(\x05\x12\x0f\n\x07userIDs\x18\x02 \x03(\t\"$\n\x12GetUserWorkRequest\x12\x0e\n\x06userID\x18\x01 \x01(\t\"#\n\x13GetUserWorkResponse\x12\x0c\n\x04work\x18\x01 \x01(\x05\x32\xfe\x03\n\tDatastore\x12H\n\rAddUserResult\x12\x1f.datastore.AddUserResultRequest\x1a\x16.google.protobuf.Empty\x12P\n\x11SetScenarioResult\x12#.datastore.SetScenarioResultRequest\x1a\x16.google.protobuf.Empty\x12X\n\x0fMoveUserResults\x12!.datastore.MoveUserResultsRequest\x1a\".datastore.MoveUserResultsResponse\x12\x61\n\x12MoveScenarioResult\x12$.datastore.MoveScenarioResultRequest\x1a%.datastore.MoveScenarioResultResponse\x12J\n\x0e\x44istributeWork\x12 .datastore.DistributeWorkRequest\x1a\x16.google.protobuf.Empty\x12L\n\x0bGetUserWork\x12\x1d.datastore.GetUserWorkRequest\x1a\x1e.datastore.GetUserWorkResponseB/Z-github.com/cicadatesting/datastore-client/apib\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_empty__pb2.DESCRIPTOR,google_dot_protobuf_dot_wrappers__pb2.DESCRIPTOR,])




_ADDUSERRESULTREQUEST = _descriptor.Descriptor(
  name='AddUserResultRequest',
  full_name='datastore.AddUserResultRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='userID', full_name='datastore.AddUserResultRequest.userID', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='result', full_name='datastore.AddUserResultRequest.result', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=106,
  serialized_end=160,
)


_SETSCENARIORESULTREQUEST = _descriptor.Descriptor(
  name='SetScenarioResultRequest',
  full_name='datastore.SetScenarioResultRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='scenarioID', full_name='datastore.SetScenarioResultRequest.scenarioID', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='output', full_name='datastore.SetScenarioResultRequest.output', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='exception', full_name='datastore.SetScenarioResultRequest.exception', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='logs', full_name='datastore.SetScenarioResultRequest.logs', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timeTaken', full_name='datastore.SetScenarioResultRequest.timeTaken', index=4,
      number=5, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=163,
  serialized_end=337,
)


_MOVEUSERRESULTSREQUEST = _descriptor.Descriptor(
  name='MoveUserResultsRequest',
  full_name='datastore.MoveUserResultsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='userIDs', full_name='datastore.MoveUserResultsRequest.userIDs', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=339,
  serialized_end=380,
)


_MOVEUSERRESULTSRESPONSE = _descriptor.Descriptor(
  name='MoveUserResultsResponse',
  full_name='datastore.MoveUserResultsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='results', full_name='datastore.MoveUserResultsResponse.results', index=0,
      number=1, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=382,
  serialized_end=424,
)


_MOVESCENARIORESULTREQUEST = _descriptor.Descriptor(
  name='MoveScenarioResultRequest',
  full_name='datastore.MoveScenarioResultRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='scenarioID', full_name='datastore.MoveScenarioResultRequest.scenarioID', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=426,
  serialized_end=473,
)


_MOVESCENARIORESULTRESPONSE = _descriptor.Descriptor(
  name='MoveScenarioResultResponse',
  full_name='datastore.MoveScenarioResultResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='datastore.MoveScenarioResultResponse.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='output', full_name='datastore.MoveScenarioResultResponse.output', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='exception', full_name='datastore.MoveScenarioResultResponse.exception', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='logs', full_name='datastore.MoveScenarioResultResponse.logs', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='datastore.MoveScenarioResultResponse.timestamp', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timeTaken', full_name='datastore.MoveScenarioResultResponse.timeTaken', index=5,
      number=6, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=476,
  serialized_end=663,
)


_DISTRIBUTEWORKREQUEST = _descriptor.Descriptor(
  name='DistributeWorkRequest',
  full_name='datastore.DistributeWorkRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='work', full_name='datastore.DistributeWorkRequest.work', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='userIDs', full_name='datastore.DistributeWorkRequest.userIDs', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=665,
  serialized_end=719,
)


_GETUSERWORKREQUEST = _descriptor.Descriptor(
  name='GetUserWorkRequest',
  full_name='datastore.GetUserWorkRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='userID', full_name='datastore.GetUserWorkRequest.userID', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=721,
  serialized_end=757,
)


_GETUSERWORKRESPONSE = _descriptor.Descriptor(
  name='GetUserWorkResponse',
  full_name='datastore.GetUserWorkResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='work', full_name='datastore.GetUserWorkResponse.work', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=759,
  serialized_end=794,
)

_SETSCENARIORESULTREQUEST.fields_by_name['output'].message_type = google_dot_protobuf_dot_wrappers__pb2._STRINGVALUE
_SETSCENARIORESULTREQUEST.fields_by_name['exception'].message_type = google_dot_protobuf_dot_wrappers__pb2._STRINGVALUE
_MOVESCENARIORESULTRESPONSE.fields_by_name['output'].message_type = google_dot_protobuf_dot_wrappers__pb2._STRINGVALUE
_MOVESCENARIORESULTRESPONSE.fields_by_name['exception'].message_type = google_dot_protobuf_dot_wrappers__pb2._STRINGVALUE
DESCRIPTOR.message_types_by_name['AddUserResultRequest'] = _ADDUSERRESULTREQUEST
DESCRIPTOR.message_types_by_name['SetScenarioResultRequest'] = _SETSCENARIORESULTREQUEST
DESCRIPTOR.message_types_by_name['MoveUserResultsRequest'] = _MOVEUSERRESULTSREQUEST
DESCRIPTOR.message_types_by_name['MoveUserResultsResponse'] = _MOVEUSERRESULTSRESPONSE
DESCRIPTOR.message_types_by_name['MoveScenarioResultRequest'] = _MOVESCENARIORESULTREQUEST
DESCRIPTOR.message_types_by_name['MoveScenarioResultResponse'] = _MOVESCENARIORESULTRESPONSE
DESCRIPTOR.message_types_by_name['DistributeWorkRequest'] = _DISTRIBUTEWORKREQUEST
DESCRIPTOR.message_types_by_name['GetUserWorkRequest'] = _GETUSERWORKREQUEST
DESCRIPTOR.message_types_by_name['GetUserWorkResponse'] = _GETUSERWORKRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

AddUserResultRequest = _reflection.GeneratedProtocolMessageType('AddUserResultRequest', (_message.Message,), {
  'DESCRIPTOR' : _ADDUSERRESULTREQUEST,
  '__module__' : 'cicadad.protos.datastore_pb2'
  # @@protoc_insertion_point(class_scope:datastore.AddUserResultRequest)
  })
_sym_db.RegisterMessage(AddUserResultRequest)

SetScenarioResultRequest = _reflection.GeneratedProtocolMessageType('SetScenarioResultRequest', (_message.Message,), {
  'DESCRIPTOR' : _SETSCENARIORESULTREQUEST,
  '__module__' : 'cicadad.protos.datastore_pb2'
  # @@protoc_insertion_point(class_scope:datastore.SetScenarioResultRequest)
  })
_sym_db.RegisterMessage(SetScenarioResultRequest)

MoveUserResultsRequest = _reflection.GeneratedProtocolMessageType('MoveUserResultsRequest', (_message.Message,), {
  'DESCRIPTOR' : _MOVEUSERRESULTSREQUEST,
  '__module__' : 'cicadad.protos.datastore_pb2'
  # @@protoc_insertion_point(class_scope:datastore.MoveUserResultsRequest)
  })
_sym_db.RegisterMessage(MoveUserResultsRequest)

MoveUserResultsResponse = _reflection.GeneratedProtocolMessageType('MoveUserResultsResponse', (_message.Message,), {
  'DESCRIPTOR' : _MOVEUSERRESULTSRESPONSE,
  '__module__' : 'cicadad.protos.datastore_pb2'
  # @@protoc_insertion_point(class_scope:datastore.MoveUserResultsResponse)
  })
_sym_db.RegisterMessage(MoveUserResultsResponse)

MoveScenarioResultRequest = _reflection.GeneratedProtocolMessageType('MoveScenarioResultRequest', (_message.Message,), {
  'DESCRIPTOR' : _MOVESCENARIORESULTREQUEST,
  '__module__' : 'cicadad.protos.datastore_pb2'
  # @@protoc_insertion_point(class_scope:datastore.MoveScenarioResultRequest)
  })
_sym_db.RegisterMessage(MoveScenarioResultRequest)

MoveScenarioResultResponse = _reflection.GeneratedProtocolMessageType('MoveScenarioResultResponse', (_message.Message,), {
  'DESCRIPTOR' : _MOVESCENARIORESULTRESPONSE,
  '__module__' : 'cicadad.protos.datastore_pb2'
  # @@protoc_insertion_point(class_scope:datastore.MoveScenarioResultResponse)
  })
_sym_db.RegisterMessage(MoveScenarioResultResponse)

DistributeWorkRequest = _reflection.GeneratedProtocolMessageType('DistributeWorkRequest', (_message.Message,), {
  'DESCRIPTOR' : _DISTRIBUTEWORKREQUEST,
  '__module__' : 'cicadad.protos.datastore_pb2'
  # @@protoc_insertion_point(class_scope:datastore.DistributeWorkRequest)
  })
_sym_db.RegisterMessage(DistributeWorkRequest)

GetUserWorkRequest = _reflection.GeneratedProtocolMessageType('GetUserWorkRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETUSERWORKREQUEST,
  '__module__' : 'cicadad.protos.datastore_pb2'
  # @@protoc_insertion_point(class_scope:datastore.GetUserWorkRequest)
  })
_sym_db.RegisterMessage(GetUserWorkRequest)

GetUserWorkResponse = _reflection.GeneratedProtocolMessageType('GetUserWorkResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETUSERWORKRESPONSE,
  '__module__' : 'cicadad.protos.datastore_pb2'
  # @@protoc_insertion_point(class_scope:datastore.GetUserWorkResponse)
  })
_sym_db.RegisterMessage(GetUserWorkResponse)


DESCRIPTOR._options = None

_DATASTORE = _descriptor.ServiceDescriptor(
  name='Datastore',
  full_name='datastore.Datastore',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=797,
  serialized_end=1307,
  methods=[
  _descriptor.MethodDescriptor(
    name='AddUserResult',
    full_name='datastore.Datastore.AddUserResult',
    index=0,
    containing_service=None,
    input_type=_ADDUSERRESULTREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='SetScenarioResult',
    full_name='datastore.Datastore.SetScenarioResult',
    index=1,
    containing_service=None,
    input_type=_SETSCENARIORESULTREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='MoveUserResults',
    full_name='datastore.Datastore.MoveUserResults',
    index=2,
    containing_service=None,
    input_type=_MOVEUSERRESULTSREQUEST,
    output_type=_MOVEUSERRESULTSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='MoveScenarioResult',
    full_name='datastore.Datastore.MoveScenarioResult',
    index=3,
    containing_service=None,
    input_type=_MOVESCENARIORESULTREQUEST,
    output_type=_MOVESCENARIORESULTRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DistributeWork',
    full_name='datastore.Datastore.DistributeWork',
    index=4,
    containing_service=None,
    input_type=_DISTRIBUTEWORKREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetUserWork',
    full_name='datastore.Datastore.GetUserWork',
    index=5,
    containing_service=None,
    input_type=_GETUSERWORKREQUEST,
    output_type=_GETUSERWORKRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_DATASTORE)

DESCRIPTOR.services_by_name['Datastore'] = _DATASTORE

# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cicadad/protos/container_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='cicadad/protos/container_service.proto',
  package='container_service',
  syntax='proto3',
  serialized_options=b'Z.github.com/cicadatesting/container-service/api',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n&cicadad/protos/container_service.proto\x12\x11\x63ontainer_service\x1a\x1bgoogle/protobuf/empty.proto\x1a\x1egoogle/protobuf/wrappers.proto\"\xb8\x03\n\x15StartContainerRequest\x12\r\n\x05image\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0f\n\x07\x63ommand\x18\x03 \x03(\t\x12\x0e\n\x06labels\x18\x04 \x03(\t\x12>\n\x03\x65nv\x18\x05 \x03(\x0b\x32\x31.container_service.StartContainerRequest.EnvEntry\x12\x30\n\x07volumes\x18\x06 \x03(\x0b\x32\x1f.container_service.DockerVolume\x12-\n\x08hostPort\x18\x07 \x01(\x0b\x32\x1b.google.protobuf.Int32Value\x12\x32\n\rcontainerPort\x18\x08 \x01(\x0b\x32\x1b.google.protobuf.Int32Value\x12-\n\x07network\x18\t \x01(\x0b\x32\x1c.google.protobuf.StringValue\x12\x31\n\rcreateNetwork\x18\n \x01(\x0b\x32\x1a.google.protobuf.BoolValue\x1a*\n\x08\x45nvEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"3\n\x0c\x44ockerVolume\x12\x0e\n\x06source\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65stination\x18\x02 \x01(\t\"+\n\x14StopContainerRequest\x12\x13\n\x0b\x63ontainerID\x18\x01 \x01(\t2\xb8\x01\n\x10\x43ontainerService\x12R\n\x0eStartContainer\x12(.container_service.StartContainerRequest\x1a\x16.google.protobuf.Empty\x12P\n\rStopContainer\x12\'.container_service.StopContainerRequest\x1a\x16.google.protobuf.EmptyB0Z.github.com/cicadatesting/container-service/apib\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_empty__pb2.DESCRIPTOR,google_dot_protobuf_dot_wrappers__pb2.DESCRIPTOR,])




_STARTCONTAINERREQUEST_ENVENTRY = _descriptor.Descriptor(
  name='EnvEntry',
  full_name='container_service.StartContainerRequest.EnvEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='container_service.StartContainerRequest.EnvEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='container_service.StartContainerRequest.EnvEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=521,
  serialized_end=563,
)

_STARTCONTAINERREQUEST = _descriptor.Descriptor(
  name='StartContainerRequest',
  full_name='container_service.StartContainerRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='image', full_name='container_service.StartContainerRequest.image', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='container_service.StartContainerRequest.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='command', full_name='container_service.StartContainerRequest.command', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='labels', full_name='container_service.StartContainerRequest.labels', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='env', full_name='container_service.StartContainerRequest.env', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='volumes', full_name='container_service.StartContainerRequest.volumes', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='hostPort', full_name='container_service.StartContainerRequest.hostPort', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='containerPort', full_name='container_service.StartContainerRequest.containerPort', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='network', full_name='container_service.StartContainerRequest.network', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='createNetwork', full_name='container_service.StartContainerRequest.createNetwork', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_STARTCONTAINERREQUEST_ENVENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=123,
  serialized_end=563,
)


_DOCKERVOLUME = _descriptor.Descriptor(
  name='DockerVolume',
  full_name='container_service.DockerVolume',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='source', full_name='container_service.DockerVolume.source', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='destination', full_name='container_service.DockerVolume.destination', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=565,
  serialized_end=616,
)


_STOPCONTAINERREQUEST = _descriptor.Descriptor(
  name='StopContainerRequest',
  full_name='container_service.StopContainerRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='containerID', full_name='container_service.StopContainerRequest.containerID', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=618,
  serialized_end=661,
)

_STARTCONTAINERREQUEST_ENVENTRY.containing_type = _STARTCONTAINERREQUEST
_STARTCONTAINERREQUEST.fields_by_name['env'].message_type = _STARTCONTAINERREQUEST_ENVENTRY
_STARTCONTAINERREQUEST.fields_by_name['volumes'].message_type = _DOCKERVOLUME
_STARTCONTAINERREQUEST.fields_by_name['hostPort'].message_type = google_dot_protobuf_dot_wrappers__pb2._INT32VALUE
_STARTCONTAINERREQUEST.fields_by_name['containerPort'].message_type = google_dot_protobuf_dot_wrappers__pb2._INT32VALUE
_STARTCONTAINERREQUEST.fields_by_name['network'].message_type = google_dot_protobuf_dot_wrappers__pb2._STRINGVALUE
_STARTCONTAINERREQUEST.fields_by_name['createNetwork'].message_type = google_dot_protobuf_dot_wrappers__pb2._BOOLVALUE
DESCRIPTOR.message_types_by_name['StartContainerRequest'] = _STARTCONTAINERREQUEST
DESCRIPTOR.message_types_by_name['DockerVolume'] = _DOCKERVOLUME
DESCRIPTOR.message_types_by_name['StopContainerRequest'] = _STOPCONTAINERREQUEST
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

StartContainerRequest = _reflection.GeneratedProtocolMessageType('StartContainerRequest', (_message.Message,), {

  'EnvEntry' : _reflection.GeneratedProtocolMessageType('EnvEntry', (_message.Message,), {
    'DESCRIPTOR' : _STARTCONTAINERREQUEST_ENVENTRY,
    '__module__' : 'cicadad.protos.container_service_pb2'
    # @@protoc_insertion_point(class_scope:container_service.StartContainerRequest.EnvEntry)
    })
  ,
  'DESCRIPTOR' : _STARTCONTAINERREQUEST,
  '__module__' : 'cicadad.protos.container_service_pb2'
  # @@protoc_insertion_point(class_scope:container_service.StartContainerRequest)
  })
_sym_db.RegisterMessage(StartContainerRequest)
_sym_db.RegisterMessage(StartContainerRequest.EnvEntry)

DockerVolume = _reflection.GeneratedProtocolMessageType('DockerVolume', (_message.Message,), {
  'DESCRIPTOR' : _DOCKERVOLUME,
  '__module__' : 'cicadad.protos.container_service_pb2'
  # @@protoc_insertion_point(class_scope:container_service.DockerVolume)
  })
_sym_db.RegisterMessage(DockerVolume)

StopContainerRequest = _reflection.GeneratedProtocolMessageType('StopContainerRequest', (_message.Message,), {
  'DESCRIPTOR' : _STOPCONTAINERREQUEST,
  '__module__' : 'cicadad.protos.container_service_pb2'
  # @@protoc_insertion_point(class_scope:container_service.StopContainerRequest)
  })
_sym_db.RegisterMessage(StopContainerRequest)


DESCRIPTOR._options = None
_STARTCONTAINERREQUEST_ENVENTRY._options = None

_CONTAINERSERVICE = _descriptor.ServiceDescriptor(
  name='ContainerService',
  full_name='container_service.ContainerService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=664,
  serialized_end=848,
  methods=[
  _descriptor.MethodDescriptor(
    name='StartContainer',
    full_name='container_service.ContainerService.StartContainer',
    index=0,
    containing_service=None,
    input_type=_STARTCONTAINERREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='StopContainer',
    full_name='container_service.ContainerService.StopContainer',
    index=1,
    containing_service=None,
    input_type=_STOPCONTAINERREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_CONTAINERSERVICE)

DESCRIPTOR.services_by_name['ContainerService'] = _CONTAINERSERVICE

# @@protoc_insertion_point(module_scope)



from typing import Any, Callable, Dict, List, Optional, Set
from datetime import datetime, timedelta
import uuid
import time
import json
import io
import traceback

from pydantic import BaseModel, Field

from cicadad.util.context import encode_context
from cicadad.core import containers
from cicadad.services import datastore, container_service
from cicadad.util import printing


class UserCommands(object):
    def __init__(
        self,
        scenario: "Scenario",
        user_id: str,
        datastore_address: str,
    ):
        """Commands available to user functions.

        Args:
            scenario (Scenario): Scenario being run
            user_id (str): ID of current user
            scenario_id (str): ID of current scenario
            datastore_address: Address of datastore client
        """
        self.scenario = scenario
        self.user_id = user_id
        self.datastore_address = datastore_address

        self.available_work = 0

    def is_up(self):
        """Check if user is still running

        Returns:
            bool: User is up
        """
        # HACK: User is stopped by scenario, but may be useful to explicitly
        # tell user to stop to allow for a graceful user-defined exit
        return True

    def has_work(self, timeout_ms: Optional[int] = 1000):
        """Check if user has remaining invocations

        Args:
            timeout_ms (int, optional): Time to wait for work event to appear before returning. Defaults to 1000.

        Returns:
            bool: User has work
        """
        if self.available_work < 1:
            self.available_work = datastore.get_work(
                self.user_id, self.datastore_address
            )

            if self.available_work < 1 and timeout_ms is not None:
                time.sleep(timeout_ms / 1000)
                self.available_work = datastore.get_work(
                    self.user_id, self.datastore_address
                )

        has_available_work = self.available_work > 0

        if has_available_work:
            self.available_work -= 1

        return has_available_work

    def run(self, *args, log_traceback=True, **kwargs):
        """Run scenario function with arguments; capture exception and logs

        Args:
            log_traceback (bool, optional): Print out traceback for exception. Defaults to True.

        Returns:
            Tuple[Any, Exception, str]: Output, exception, and logs captured
        """
        buffer = io.StringIO()

        with printing.stdout_redirect(buffer):
            try:
                output, exception = self.scenario.fn(*args, **kwargs), None
            except Exception as e:
                output, exception = None, e

                if log_traceback:
                    print("Exception traceback:")
                    traceback.print_tb(e.__traceback__)

        return output, exception, buffer.getvalue()

    def report_result(
        self, output: Any, exception: Any, logs: Optional[str], time_taken: int
    ):
        """Report result for scenario invocation from user to scenario

        Args:
            output (Any): Function output
            exception (Any): Function exception
            logs (Optional[str]): Function logs
            time_taken (int): Time taken in seconds to call function once
        """
        result = datastore.Result(
            id=str(uuid.uuid4()),
            output=output,
            exception=exception,
            logs=logs,
            timestamp=datetime.now(),
            time_taken=time_taken,
        )

        datastore.add_user_result(self.user_id, result, self.datastore_address)


class ScenarioCommands(object):
    def __init__(
        self,
        scenario: "Scenario",
        image: str,
        network: str,
        scenario_id: str,
        datastore_address: str,
        container_service_address: str,
        context: dict,
    ):
        """Commands available to a scenario

        Args:
            scenario (Scenario): Scenario being run
            image (str): Docker image for scenario
            network (str): Docker network for scenario
            scenario_id (str): ID of scenario run
            datastore_address (str): Address of datastore to pass to users
            container_service (str): Address of container service to start and stop users
            context (dict): Context data to pass to users
        """
        self.scenario = scenario
        self.image = image
        self.network = network
        self.datastore_address = datastore_address
        self.container_service_address = container_service_address
        self.context = context

        self.scenario_id = scenario_id
        self.user_ids: Set[str] = set()
        self.num_users = 0
        self.buffered_work = 0
        self.aggregated_results = None
        self.errors: List[str] = []

    def scale_users(self, n: int):
        """Change number of running users

        Args:
            n (int): Desired number of users
        """
        if n > self.num_users:
            self.start_users(n - self.num_users)
        else:
            self.stop_users(self.num_users - n)

    def start_users(self, n: int):
        """Start users for scenario

        Args:
            n (int): Number of users to start
        """
        # FEATURE: will need to determine whether docker or kube args
        # FEATURE: attach additional container args to scenario
        if n < 0:
            raise ValueError("Must supply a positive number of users to start")

        if n == 0:
            return

        # FEATURE: mount env and volumes for user
        for _ in range(n):
            user_id = f"user-{self.scenario.name}-{str(uuid.uuid4())[:8]}"
            encoded_context = encode_context(self.context)

            args = containers.DockerServerArgs(
                image=self.image,
                name=user_id,
                command=[
                    "run-user",
                    "--name",
                    self.scenario.name,
                    "--user-id",
                    user_id,
                    "--datastore-address",
                    self.datastore_address,
                    "--encoded-context",
                    encoded_context,
                ],
                labels=["cicada-distributed-user", self.scenario.name],
                # env: Dict[str, str]={}
                # volumes: Optional[List[Volume]]
                # host_port: Optional[int]
                # container_port: Optional[int]
                network=self.network,
                # create_network: bool=True
            )

            self.user_ids.add(user_id)
            container_service.start_container(args, self.container_service_address)

        self.num_users += n

        # If called add_work before start_users, flush the saved work to the
        # new group of users
        if self.buffered_work > 0:
            self.add_work(self.buffered_work)
            self.buffered_work = 0

    def stop_users(self, n: int):
        """Stop a given number of users

        Args:
            n (int): Number of users to stop
        """
        if self.num_users < n:
            raise ValueError(f"Scenario currently has less than {n} users")

        if n < 0:
            raise ValueError("Must supply a positive number of users to stop")

        if n == 0:
            return

        remaining = n

        for user_id in self.user_ids.copy():
            if remaining < 1:
                break

            container_service.stop_container(user_id, self.container_service_address)
            self.user_ids.remove(user_id)
            self.num_users -= 1
            remaining -= 1

    def add_work(self, n: int):
        """Distribute work to all users in scenario

        Args:
            n (int): Amount of work to distribute across user pool
        """
        # If no users exist, save in a buffer
        if self.user_ids == set():
            self.buffered_work += n
            return

        datastore.distribute_work(n, list(self.user_ids), self.datastore_address)

    def get_latest_results(
        self,
        timeout_ms: Optional[int] = 1000,
    ):
        """Gathers results produced by users

        Args:
            timeout_ms (int, optional): Time to wait for results. Defaults to 1000.

        Returns:
            List[Result]: List of latest results collected
        """
        results = datastore.move_user_results(self.user_ids, self.datastore_address)

        if results == [] and timeout_ms is not None:
            time.sleep(timeout_ms / 1000)
            results = datastore.move_user_results(self.user_ids, self.datastore_address)

        return results

    def aggregate_results(self, latest_results: List[datastore.Result]) -> Any:
        """Run scenario aggregator function against latest gathered results and
        save aggregate

        Args:
            latest_results (List[Result]): Results to run aggregator function on

        Returns:
            Any: Result of scenario aggregator function
        """
        if self.scenario.result_aggregator is not None:
            self.aggregated_results = self.scenario.result_aggregator(
                self.aggregated_results, latest_results
            )
        elif latest_results != []:
            # Default aggregator when results are not empty
            self.aggregated_results = latest_results[-1].output

        return self.aggregated_results

    def verify_results(
        self, latest_results: List[datastore.Result]
    ) -> Optional[List[str]]:
        """Run scenario result verification function against latest results

        Args:
            latest_results (List[Result]): Last results to be collected

        Returns:
            Optional[List[str]]: List of error strings gathered for scenario
        """
        if self.scenario.result_verifier is not None:
            errors = self.scenario.result_verifier(latest_results)

            self.errors.extend(errors)
            return errors

        return None

    # FEATURE: get number of healthy users in scenario, healthy users per group

    # def create_metric(
    #     name: str, description: Union[str, int, float], type: str = "HISTOGRAM"
    # ):
    #     # NOTE: name must be prometheus allowed
    #     return

    # def put_metric(name: str, value: str, description: str, type: str = "HISTOGRAM"):
    #     # create metric if not created yet, then report metric
    #     return

    # def report_metric(name: str, value: str):
    #     return


def filter_scenarios_by_tag(scenarios: List["Scenario"], tags: List[str]):
    """Filter scenarios that have tags in the list of tags provided. Returns
    all scenarios if tag list is empty

    Args:
        scenarios (List[Scenario]): List of scenarios
        tags (List[str]): List of tags to filter scenarios with

    Returns:
        List[Scenario]: List of filtered scenarios
    """
    if tags == []:
        return scenarios

    return [s for s in scenarios if set(s.tags).intersection(set(tags)) != set()]


class TestStatus(BaseModel):
    type: str
    scenario: str
    message: str
    context: str


def test_runner(
    scenarios: List["Scenario"],
    image: str,
    network: str,
    datastore_address: str,
    container_service_address: str,
):
    started: Dict[str, str] = {}
    results: Dict[str, dict] = {}

    # Start scenarios with no dependencies
    for scenario in scenarios:
        if scenario.dependencies == []:
            # FIXME: move to function
            scenario_id = str(uuid.uuid4())[:8]
            container_id = f"scenario-{scenario.name}-{scenario_id}"
            encoded_context = encode_context(results)

            args = containers.DockerServerArgs(
                image=image,
                name=container_id,
                command=[
                    "run-scenario",
                    "--name",
                    scenario.name,
                    "--scenario-id",
                    scenario_id,
                    "--image",
                    image,
                    "--network",
                    network,
                    "--encoded-context",
                    encoded_context,
                    "--datastore-address",
                    datastore_address,
                    "--container-service-address",
                    container_service_address,
                ],
                labels=["cicada-distributed-scenario", scenario.name],
                # env: Dict[str, str]={}
                # volumes: Optional[List[Volume]]
                # host_port: Optional[int]
                # container_port: Optional[int]
                network=network,
                # create_network: bool=True
            )

            container_service.start_container(args, container_service_address)
            started[scenario.name] = scenario_id

            yield TestStatus(
                type="SCENARIO_STARTED",
                scenario=scenario.name,
                message=f"Started scenario: {scenario.name} ({container_id})",
                context=json.dumps(results),
            )

    # listen to completed events and start scenarios with dependencies
    while len(results) != len(scenarios):
        for scenario_name in [sn for sn in started if sn not in results]:
            # FIXME: move logic to eventing function
            # FEATURE: stream back metrics gathered from scenarios
            result = datastore.move_scenario_result(
                started[scenario_name],
                datastore_address,
            )

            if result is not None:
                results[scenario_name] = result

                yield TestStatus(
                    type="SCENARIO_FINISHED",
                    scenario=scenario_name,
                    message=f"Finished Scenario: {scenario_name}",
                    context=json.dumps(results),
                )

        for scenario in [s for s in scenarios if s.name not in started]:
            # FIXME: move filtering to function
            if all(
                dep.name in results and results[dep.name]["exception"] is None
                for dep in scenario.dependencies
            ):
                scenario_id = str(uuid.uuid4())[:8]
                container_id = f"scenario-{scenario.name}-{scenario_id}"
                encoded_context = encode_context(results)

                args = containers.DockerServerArgs(
                    image=image,
                    name=container_id,
                    command=[
                        "run-scenario",
                        "--name",
                        scenario.name,
                        "--scenario-id",
                        scenario_id,
                        "--image",
                        image,
                        "--network",
                        network,
                        "--encoded-context",
                        encoded_context,
                        "--datastore-address",
                        datastore_address,
                        "--container-service-address",
                        container_service_address,
                    ],
                    labels=["cicada-distributed-scenario", scenario.name],
                    # env: Dict[str, str]={}
                    # volumes: Optional[List[Volume]]
                    # host_port: Optional[int]
                    # container_port: Optional[int]
                    network=network,
                    # create_network: bool=True
                )

                container_service.start_container(args, container_service_address)
                started[scenario.name] = scenario_id

                yield TestStatus(
                    type="SCENARIO_STARTED",
                    scenario=scenario.name,
                    message=f"Started scenario: {scenario.name} ({container_id})",
                    context=json.dumps(results),
                )
            elif all(dep.name in results for dep in scenario.dependencies):
                # has all dependencies but some are failed
                started[scenario.name] = str(uuid.uuid4())[:8]
                results[scenario.name] = json.loads(
                    datastore.Result(
                        id=str(uuid.uuid4()),
                        output=None,
                        exception="Skipped",
                        logs="",
                        timestamp=datetime.now(),
                    ).json()
                )

                yield TestStatus(
                    type="SCENARIO_FINISHED",
                    scenario=scenario.name,
                    message=f"Skipped Scenario: {scenario.name}",
                    context=json.dumps(results),
                )

        # NOTE: maybe make this configurable or shorter?
        time.sleep(1)


def scenario_runner(
    scenario: "Scenario",
    image: str,
    network: str,
    scenario_id: str,
    datastore_address: str,
    container_service_address: str,
    context: dict,
):
    """Set up scenario environment and run scenario. Capture output and exceptions

    Args:
        scenario (Scenario): Scenario being run
        image (str): Image for scenario
        network (str): Network to run scenario containers in
        scenario_id (str): ID generated for scenario run
        datastore_address (str): Address of datastore passed to users
        container_service (str): Address of container service to start and stop users
        context (dict): Test context to pass to users
    """
    scenario_commands = ScenarioCommands(
        scenario,
        image,
        network,
        scenario_id,
        datastore_address,
        container_service_address,
        context,
    )

    buffer = io.StringIO()

    start = datetime.now()
    output: Optional[Any] = None
    exception: Optional[Exception] = None

    with printing.stdout_redirect(buffer):
        try:
            scenario.load_model(scenario_commands, context)  # type: ignore

            if scenario.output_transformer is not None:
                output = scenario.output_transformer(
                    scenario_commands.aggregated_results
                )
            else:
                output = scenario_commands.aggregated_results

            if scenario_commands.errors != []:
                error_strs = [
                    f"{len(scenario_commands.errors)} error(s) were raised in scenario {scenario.name}:"
                ] + scenario_commands.errors

                exception = AssertionError("\n".join(error_strs))
            else:
                exception = None
        except Exception as e:
            output, exception = None, e

    end = datetime.now()

    # NOTE: this doesn't get reported to the user, just stays in container logs
    if exception is not None:
        print(exception)
        traceback.print_tb(exception.__traceback__)

    # Clean up
    scenario_commands.scale_users(0)

    datastore.set_scenario_result(
        scenario_id=scenario_id,
        output=output,
        exception=exception,
        logs=buffer.getvalue(),
        time_taken=(end - start).seconds,
        address=datastore_address,
    )


def user_runner(
    scenario: "Scenario",
    user_id: str,
    datastore_address: str,
    context: dict,
):
    """Set up environment for user and run user

    Args:
        scenario (Scenario): Scenario being run
        user_id (str): ID generated for user
        datastore_address (str): Address of datastore client to save results
        context (dict): Test context containing previous scenario results
    """
    user_commands = UserCommands(
        scenario,
        user_id,
        datastore_address,
    )

    scenario.user_loop(user_commands, context)  # type: ignore


# FIXME: move these to another module
def while_has_work(polling_timeout_ms: int = 1000):
    """Run user if work is available or continue polling

    Args:
        polling_timeout_ms (UserCommands): Time to wait for work before cycling
    """

    def closure(user_commands: UserCommands, context: dict):
        while user_commands.is_up():
            if user_commands.has_work(polling_timeout_ms):
                start = datetime.now()
                output, exception, logs = user_commands.run(context=context)
                end = datetime.now()
                user_commands.report_result(
                    output,
                    exception,
                    logs,
                    time_taken=(end - start).seconds,
                )

    return closure


def while_alive():
    """Run user if hasn't been shut down yet

    Args:
        polling_timeout_ms (UserCommands): Time to wait for work before cycling
    """

    def closure(user_commands: UserCommands, context: dict):
        while user_commands.is_up():
            start = datetime.now()
            output, exception, logs = user_commands.run(context=context)
            end = datetime.now()
            user_commands.report_result(
                output,
                exception,
                logs,
                time_taken=(end - start).seconds,
            )

    return closure


def iterations_per_second_limited(limit: int):
    """Allows a user to run a limited number of iterations per second.

    Args:
        limit (int): Max iterations per second for user
    """

    def closure(user_commands: UserCommands, context: dict):
        remaining_iterations = limit
        second_start_time = datetime.now()

        while user_commands.is_up():
            if remaining_iterations > 0:
                start = datetime.now()

                output, exception, logs = user_commands.run(context=context)

                end = datetime.now()

                remaining_iterations -= 1
                user_commands.report_result(
                    output,
                    exception,
                    logs,
                    time_taken=(end - start).seconds,
                )
            else:
                td = (second_start_time + timedelta(seconds=1)) - datetime.now()

                time.sleep(td.seconds + (td.microseconds / 1000000))

            if datetime.now() >= second_start_time + timedelta(seconds=1):
                remaining_iterations = limit
                second_start_time = datetime.now()

    return closure


def n_iterations(
    iterations: int,
    users: int,
    wait_period: int = 1,
    timeout: Optional[int] = 15,
    skip_scaledown: bool = False,
):
    """Create a load model where a pool of users is called n times.

    Args:
        iterations (int): Number of shared iterations for users to run
        users (int): Size of user pool
        wait_period (int, optional): Time in seconds to between polling for results. Defaults to 1.
        timeout (Optional[int], optional): Time in seconds for scenario to complete before failing. Defaults to 15.
        skip_scaledown (bool): Skip scaledown of users after running load function
    """

    def closure(scenario_commands: ScenarioCommands, context: dict):
        scenario_commands.scale_users(users)
        scenario_commands.add_work(iterations)

        # wait for completion
        num_results = 0
        start_time = datetime.now()

        while num_results < iterations:
            if (
                timeout is not None
                and timeout > 0
                and datetime.now() > start_time + timedelta(seconds=timeout)
            ):
                scenario_commands.scale_users(0)
                raise AssertionError("Timed out waiting for results")

            latest_results = scenario_commands.get_latest_results()

            scenario_commands.aggregate_results(latest_results)
            scenario_commands.verify_results(latest_results)
            num_results += len(latest_results)

            time.sleep(wait_period)

        if skip_scaledown:
            return

        scenario_commands.scale_users(0)

    return closure


def run_scenario_once(wait_period: int = 1, timeout: Optional[int] = 15):
    """Run scenario one time with one user.

    Args:
        wait_period (int, optional): Time in seconds to wait before polling for results. Defaults to 1.
        timeout (int, optional): Time in seconds to wait for scenario to complete before failing. Defaults to 15.

    Returns:
        Callable: Closure for configured load model
    """
    return n_iterations(1, 1, wait_period, timeout)


def n_seconds(
    seconds: int,
    users: int,
    wait_period: int = 1,
    skip_scaledown=False,
):
    """Run the scenario for a specified duration. Should be used with the
    'while_alive' user loop.

    Args:
        seconds (int): Number of seconds to run scenario
        users (int): Number of users to start for scenario
        wait_period (int, optional): Time in seconds to wait before polling for results. Defaults to 1.
        skip_scaledown (bool): Skip scaledown of users after running load function
    """

    def closure(scenario_commands: ScenarioCommands, context: dict):
        scenario_commands.scale_users(users)

        # collect results for specified seconds
        start_time = datetime.now()

        while datetime.now() < start_time + timedelta(seconds=seconds):
            # FIXME: ensure remaining results are collected
            latest_results = scenario_commands.get_latest_results()

            scenario_commands.aggregate_results(latest_results)
            scenario_commands.verify_results(latest_results)

            time.sleep(wait_period)

        if skip_scaledown:
            return

        scenario_commands.scale_users(0)

    return closure


def n_users_ramping(
    seconds: int,
    target_users: int,
    wait_period: int = 1,
    skip_scaledown: bool = True,
):
    """Scale users to target over the duration of the time specified.

    Use this to scale users smoothly.

    Args:
        seconds (int): Amount of time to spend ramping users
        target_users (int): Number of users to ramp to.
        wait_period (int, optional): Time in seconds to wait between scaling batch of users. Defaults to 1.
        skip_scaledown (bool, optional): Do not scale down users after load model completes. Defaults to True.
    """

    def closure(scenario_commands: ScenarioCommands, context: dict):
        start_time = datetime.now()
        starting_users = scenario_commands.num_users
        buffered_users = float(0)

        while datetime.now() <= start_time + timedelta(seconds=seconds):
            if starting_users > target_users:
                users_to_stop = (starting_users - target_users) / int(
                    seconds / wait_period
                )

                buffered_users += users_to_stop

                if int(buffered_users) > 0:
                    scenario_commands.stop_users(int(buffered_users))
                    buffered_users -= int(buffered_users)
            else:
                users_to_start = (target_users - starting_users) / int(
                    seconds / wait_period
                )

                buffered_users += users_to_start

                if int(buffered_users) > 0:
                    scenario_commands.start_users(int(buffered_users))
                    buffered_users -= int(buffered_users)

            latest_results = scenario_commands.get_latest_results()

            scenario_commands.aggregate_results(latest_results)
            scenario_commands.verify_results(latest_results)

            time.sleep(wait_period)

        if skip_scaledown:
            scenario_commands.scale_users(target_users)
            return

        scenario_commands.scale_users(0)

    return closure


def ramp_users_to_threshold(
    initial_users: int,
    threshold_fn: Callable[[Any], bool],
    next_users_fn: Callable[[int], int] = lambda n: n + 10,
    update_aggregate: Callable[[int, Any], Any] = lambda n, agg: f"Users: {n}",
    period_duration: int = 30,
    period_limit: Optional[int] = None,
    wait_period: int = 1,
    skip_scaledown: bool = False,
):
    """Increase number of users in scenario until a threshold based on the
    aggregated results is reached.

    Update aggregate with number of users determined by scenario.

    Args:
        initial_users (int): Users to start stage with.
        threshold_fn (Callable[[Any], bool]): Checks aggregate and returns True if threshold reached.
        next_users_fn (Callable[[int], int]): Scale number of users given current number of users.
        update_aggregate (Callable[[int, Any], Any], optional): Update scenario aggregate with result of load model.
        period_duration (int, optional): Time in seconds to wait before scaling test. Defaults to 30.
        period_limit (Optional[int], optional): Amount of scaling events before stopping stage. Defaults to None.
        wait_period (int, optional): Time in seconds to wait before polling for results. Defaults to 1.
        skip_scaledown (bool): Skip scaledown of users after running load function
    """

    def closure(scenario_commands: ScenarioCommands, context: dict):
        scenario_commands.scale_users(initial_users)
        period_count = 0
        period_start = datetime.now()

        while not threshold_fn(scenario_commands.aggregated_results) and (
            period_limit is None or period_limit < period_count
        ):
            latest_results = scenario_commands.get_latest_results()

            scenario_commands.aggregate_results(latest_results)
            scenario_commands.verify_results(latest_results)

            time.sleep(wait_period)

            if datetime.now() >= period_start + timedelta(seconds=period_duration):
                scenario_commands.scale_users(
                    next_users_fn(scenario_commands.num_users)
                )

                period_count += 1
                period_start = datetime.now()

        scenario_commands.aggregated_results = update_aggregate(
            scenario_commands.num_users, scenario_commands.aggregated_results
        )

        if skip_scaledown:
            return

        scenario_commands.scale_users(0)

    return closure


def basic_verification(latest_results: List[datastore.Result]):
    """Create error strings for each errored result.

    Format:
    * {type}: {error}

    Args:
        latest_results (List[Result]): List of results to generate errors for

    Returns:
        List[str]: List of error strings derived from results
    """
    return [
        f"* {type(result.exception)}: {str(result.exception)}"
        for result in latest_results
        if result.exception is not None
    ]


UserLoopFn = Callable[[UserCommands, dict], None]
LoadModelFn = Callable[[ScenarioCommands, dict], Any]
ResultAggregatorFn = Callable[[Optional[Any], List[datastore.Result]], Any]
ResultVerifierFn = Callable[[List[datastore.Result]], List[str]]
OutputTransformerFn = Callable[[Optional[Any]], Any]


class Scenario(BaseModel):
    name: str
    fn: Callable
    user_loop: Optional[UserLoopFn] = Field(while_has_work())
    load_model: Optional[LoadModelFn] = Field(run_scenario_once())
    dependencies: List["Scenario"] = []
    result_aggregator: Optional[ResultAggregatorFn]
    result_verifier: Optional[ResultVerifierFn] = Field(basic_verification)
    output_transformer: Optional[OutputTransformerFn]
    tags: List[str] = []


Scenario.update_forward_refs()


def load_stages(*stages: LoadModelFn):
    def closure(scenario_commands: ScenarioCommands, context: dict):
        for stage in stages:
            stage(scenario_commands, context)

        scenario_commands.scale_users(0)

    return closure


# FEATURE: signal user loop that stage has changed from scenario commands



from contextlib import contextmanager
from typing import TextIO
import io
import sys


import logging

from cicadad.util.constants import LOG_FORMAT, DATE_FORMAT


# NOTE: get log_level/log_file from env var?
def get_logger(name: str = "logger"):
    """Configure a logger with log format

    Args:
        name (str, optional): Name of logger. Defaults to "logger".

    Returns:
        Logger: Configured logger
    """
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    sh = logging.StreamHandler()
    logger.addHandler(sh)
    sh.setFormatter(logging.Formatter(LOG_FORMAT, datefmt=DATE_FORMAT))

    return logger


class CopiedStdout(io.TextIOBase):
    def __init__(self, terminal: TextIO, copy: io.TextIOBase):
        """Copy stdout for a terminal to a buffer

        Args:
            terminal (TextIO): Terminal printing stdout to (like sys.stdout)
            copy (TextIOBase): Buffer to copy stdout to
        """
        self.terminal = terminal
        self.copy = copy

    def write(self, message: str):
        """Write message to terminal and buffer

        Args:
            message (str): Message to print
        """
        self.terminal.write(message)
        self.copy.write(message)


@contextmanager
def stdout_redirect(buffer: io.TextIOBase):
    """Copy sys.stdout to a buffer while in context

    Args:
        buffer (TextIOBase): Buffer to write stdout to
    """
    old_stdout = sys.stdout
    new_stdout = CopiedStdout(old_stdout, buffer)

    try:
        sys.stdout = new_stdout  # type: ignore
        yield
    finally:
        sys.stdout = old_stdout
